# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#
import ConfigSpace
import hpbandster
import hpbandster.core.worker
import hpbandster.core.nameserver
import hpbandster.optimizers

import numpy as np
from autotune.problem import TuningProblem
from options import Options
from computer import Computer
from typing import Collection

import skopt.space
from skopt.space import *
from data import *
import math
import argparse
import functools
import time
import sys
####################################################################################################

class HpBandSterWorker(hpbandster.core.worker.Worker):

    def __init__(self, t, NS, tp, computer, niter=1, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.myworker_id = kwargs['id']
        self.tp          = tp
        self.computer    = computer
        self.t           = t
        self.NS          = NS
        self.niter       = niter
        self.count_runs  = 0
        self.timefun     = 0

    def get_configspace(self):

        class MyConstrainedConfigurationSpace(ConfigSpace.ConfigurationSpace):

            def __init__(self, tp, computer, t):

                super(MyConstrainedConfigurationSpace, self).__init__()
                self.t = t
                self.tp = tp
                self.computer = computer

            def sample_configuration(self, size=1):

                cond = False
                cpt = 0
                while (not cond):

                    cpt += 1
                    if (size == 1):
                        accepted_configurations = [super(MyConstrainedConfigurationSpace, self).sample_configuration(size=size)]
                    else:
                        accepted_configurations = super(MyConstrainedConfigurationSpace, self).sample_configuration(size=size)
                    for config in accepted_configurations:
                        t = self.t
                        print(config,type(config))
                        x = [config[p] for p in self.tp.parameter_space.dimension_names]
                        kwargs = {d.name: x[i] for (i, d) in enumerate(self.tp.parameter_space)}
                        kwargs2 = {d.name: t[i] for (i, d) in enumerate(self.tp.input_space)}
                        kwargs2.update(kwargs)
                        check_constraints = functools.partial(self.computer.evaluate_constraints, self.tp, inputs_only = False, kwargs = kwargs)
                        cond = check_constraints(kwargs2)
                        if (not cond):
                            break

                if (size == 1):
                    return accepted_configurations[0]
                else:
                    return accepted_configurations

        #config_space = ConfigSpace.ConfigurationSpace()
        config_space = MyConstrainedConfigurationSpace(self.tp, self.computer, self.t)

        for n,p in zip(self.tp.parameter_space.dimension_names,self.tp.parameter_space.dimensions):
            if (isinstance(p, Real)):
                config_space.add_hyperparameter(ConfigSpace.UniformFloatHyperparameter(n, lower = p.bounds[0], upper = p.bounds[1]))
            elif (isinstance(p, Integer)):
                config_space.add_hyperparameter(ConfigSpace.UniformIntegerHyperparameter(n, lower = p.bounds[0], upper = p.bounds[1]))
            elif (isinstance(p, Categorical)):
                config_space.add_hyperparameter(ConfigSpace.CategoricalHyperparameter(n, choices = list(p.bounds)))
            else:
                raise Exception("Unknown parameter type")
        return(config_space)

    def compute(self, config, budget, **kwargs):

        if (self.count_runs >= self.NS):

            y = float("Inf")
            state = 'OVERTIME'

            print(self.t, state)
            sys.stdout.flush()

        else:
            t1 = time.time_ns()
            t = self.t
            x = [config[p] for p in self.tp.parameter_space.dimension_names]
            kwargs = {d.name: x[i] for (i, d) in enumerate(self.tp.parameter_space)}
            kwargs2 = {d.name: t[i] for (i, d) in enumerate(self.tp.input_space)}
            kwargs2.update(kwargs)
            kwargs2['budget'] = budget
            check_constraints = functools.partial(self.computer.evaluate_constraints, self.tp, inputs_only = False, kwargs = kwargs)
            cond = check_constraints(kwargs2)

            if (cond):
                y = float(self.tp.objective(kwargs2)[0])
                state = 'OK'
            else:
                y = float("Inf")
                state = 'ERROR'
            print(t, x, y, state)
            t2 = time.time_ns()
            self.timefun=self.timefun+(t2-t1)/1e9
            sys.stdout.flush()

        self.count_runs += 1

        return({
                   'loss': y,
                   'info': {"state":state}
               })

####################################################################################################

def HpBandSter(T, NS, tp : TuningProblem, computer : Computer, run_id="HpBandSter", niter=1):

    # Initialize
    min_budget   = 1. # Minimum budget used during the optimization.
    max_budget   = 1. # Maximum budget used during the optimization.
    n_iterations = NS # Number of iterations performed by the optimizer
    n_workers    = 1  # Number of workers to run in parallel.

    X = []
    Y = []
    # Xopt = []
    # Yopt = []
    data = Data(tp)

    server = hpbandster.core.nameserver.NameServer(run_id=run_id, host='127.0.0.1', port=None)
    server.start()

    # Tune
    stats = {
        "time_total": 0,
        "time_fun": 0
    }

    timefun=0
    t1 = time.time_ns()
    print("Start HpBandSter")
    for i in range(len(T)):

        workers=[]
        for j in range(n_workers):
            w = HpBandSterWorker(t=T[i], NS=NS, tp=tp, computer=computer, niter=niter, run_id=run_id, nameserver='127.0.0.1', id=j)
            w.run(background=True)
            workers.append(w)

        bohb = hpbandster.optimizers.BOHB(configspace=workers[0].get_configspace(), run_id=run_id, nameserver='127.0.0.1', min_budget=min_budget, max_budget=max_budget)
        res = bohb.run(n_iterations=n_iterations, min_n_workers=n_workers)

        config_mapping = res.get_id2config_mapping()
        # incumbent = res.get_incumbent_id()

        xs = [[config_mapping[idx]['config'][p] for p in tp.parameter_space.dimension_names] for idx in config_mapping.keys()]
        ys = [[v['loss'] for k,v in res[idx].results.items()] for idx in config_mapping.keys()]
        # xopt = np.array([res.get_id2config_mapping()[incumbent]['config'][p] for p in tp.parameter_space.dimension_names])
        # yopt = min([v['loss'] for k,v in res[incumbent].results.items()])
        X.append(xs)
        tmp = np.array(ys).reshape((len(ys), 1))
        Y.append(tmp)
        # Xopt.append(xopt)
        # Yopt.append(yopt)
        timefun=timefun+workers[0].timefun
        bohb.shutdown(shutdown_workers=True)

    print("End HpBandSter")
    t2 = time.time_ns()
    stats['time_total'] = (t2-t1)/1e9
    stats['time_fun'] = timefun
    # Finalize

    server.shutdown()

    data.I=T
    data.P=X
    data.O=Y
    # Finalize

    return (data, stats)

def HpBandSter_bandit(T, NS, tp : TuningProblem, computer : Computer, options: Options = None, run_id="HpBandSter_bandit", niter=1):
   # Initialize
    min_budget   = options['budget_min'] # Minimum budget used during the optimization.
    max_budget   = options['budget_max'] # Maximum budget used during the optimization.
    budget_base  = options['budget_base']
    n_iterations = NS # Number of iterations performed by the optimizer
    n_workers    = 1  # Number of workers to run in parallel.
    
    X = []
    Y = []
    # Xopt = []
    # Yopt = []
    data = Data(tp)

    server = hpbandster.core.nameserver.NameServer(run_id=run_id, host='127.0.0.1', port=None)
    server.start()

    # Tune
    stats = {
        "time_total": 0,
        "time_fun": 0
    }

    timefun=0
    t1 = time.time_ns()
    for i in range(len(T)):

        workers=[]
        for j in range(n_workers):
            w = HpBandSterWorker(t=T[i], NS=NS, tp=tp, computer=computer, niter=niter, run_id=run_id, nameserver='127.0.0.1', id=j)
            w.run(background=True)
            workers.append(w)
            
        bohb = hpbandster.optimizers.BOHB(configspace=workers[0].get_configspace(), run_id=run_id, nameserver='127.0.0.1', min_budget=min_budget, max_budget=max_budget, eta=budget_base)
        res = bohb.run(n_iterations=n_iterations, min_n_workers=n_workers)

        config_mapping = res.get_id2config_mapping()

        xs = [[config_mapping[idx]['config'][p] for p in tp.parameter_space.dimension_names] for idx in config_mapping.keys()]
        ys = [[(k, v['loss']) for k,v in res[idx].results.items()] for idx in config_mapping.keys()]
        
        X.append(xs)
        tmp = np.array(ys).reshape((len(ys), 1))
        Y.append(tmp)
        timefun=timefun+workers[0].timefun
        bohb.shutdown(shutdown_workers=True)

    t2 = time.time_ns()
    stats['time_total'] = (t2-t1)/1e9	
    stats['time_fun'] = timefun
    # Finalize

    server.shutdown()

    data.I=T
    data.P=X
    data.O=Y
    # Finalize

    return (data, stats)


# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#

from autotune.problem import TuningProblem
import copy

class Problem(object):

    def __init__(self, tp : TuningProblem, driverabspath=None, models_update=None):

        self.name = tp.name

        self.IS = tp.input_space
        self.PS = tp.parameter_space
        self.OS = tp.output_space

        self.objectives   = tp.objective
        self.driverabspath   = driverabspath
        self.models_update   = models_update
        self.constraints = tp.constraints
        self.models      = tp.models
        self.constants      = tp.constants        

    @property
    def DI(self):

        return len(self.IS)

    @property
    def DP(self):

        return len(self.PS)

    @property
    def DO(self):

        return len(self.OS)




#! /usr/bin/env python

# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#
import concurrent
from concurrent import futures
import sys
import abc
from typing import Collection
import numpy as np
import scipy as sp
import functools
from joblib import *


import mpi4py

from problem import Problem
from computer import Computer
from options import Options
from data import Data
from model import Model

from pathlib import Path
import importlib
from sys import platform as _platform

class Search(abc.ABC):

    def __init__(self, problem : Problem, computer : Computer, options: Options):
        self.problem = problem
        self.computer = computer
        self.options = options

    @abc.abstractmethod
    def search(self, data : Data, models : Collection[Model], tid : int, **kwargs) -> np.ndarray:

        raise Exception("Abstract method")

    def search_multitask(self, data : Data, models : Collection[Model], tids : Collection[int] = None, i_am_manager : bool = True, **kwargs) -> Collection[np.ndarray]:

        if (tids is None):
            tids = list(range(data.NI))

        if ((kwargs['distributed_memory_parallelism'] or _platform == "darwin") and i_am_manager):   # the pgymo install on mac os seems buggy if search is not spawned 
            nproc = min(kwargs['search_multitask_processes'],data.NI)
            npernode = int(self.computer.cores/kwargs['search_multitask_threads'])
            mpi_comm = self.computer.spawn(__file__, nproc=nproc, nthreads=kwargs['search_multitask_threads'], npernode=npernode, kwargs=kwargs) # XXX add args and kwargs
            kwargs_tmp = kwargs
            if "mpi_comm" in kwargs_tmp:
                del kwargs_tmp["mpi_comm"]   # mpi_comm is not picklable
            _ = mpi_comm.bcast((self, data, models, tids, kwargs_tmp), root=mpi4py.MPI.ROOT)
            tmpdata = mpi_comm.gather(None, root=mpi4py.MPI.ROOT)
            mpi_comm.Disconnect()
            res=[]
            for p in range(int(nproc)):
                res = res + tmpdata[p]


        elif (kwargs['shared_memory_parallelism']):
            #with concurrent.futures.ProcessPoolExecutor(max_workers = kwargs['search_multitask_threads']) as executor:
            with concurrent.futures.ThreadPoolExecutor(max_workers = kwargs['search_multitask_threads']) as executor:
                # fun = functools.partial(self.search, data = data, models = models, kwargs = kwargs)
                # res = list(executor.map(fun, tids, timeout=None, chunksize=1))

                def fun(tid):
                    return self.search(data=data,models = models, tid =tid, kwargs = kwargs)
                res = list(executor.map(fun, tids, timeout=None, chunksize=1))
        else:
            fun = functools.partial(self.search, data, models, kwargs = kwargs)
            res = list(map(fun, tids))
        # print(res)
        res.sort(key = lambda x : x[0])
        return res

class SurrogateProblem(object):

    def __init__(self, problem, computer, data, models, options, tid):   # data is in the normalized space, IOrig and POrig are then generated in the original space

        self.problem = problem
        self.computer = computer
        self.data = data
        self.models = models
        self.options = options

        self.tid = tid

        self.D     = self.data.D[tid]
        self.IOrig = self.problem.IS.inverse_transform(np.array(self.data.I[tid], ndmin=2))[0]

        # self.POrig = self.data.P[tid]
        self.POrig = self.problem.PS.inverse_transform(np.array(self.data.P[tid], ndmin=2))

    def get_nobj(self):
        if(self.options['search_algo']=='pso' or self.options['search_algo']=='cmaes'):
            return 1
        else:
            return self.problem.DO

    def get_bounds(self):

        DP = self.problem.DP

        return ([0. for i in range(DP)], [1. for  i in range(DP)])

    # Acquisition function
    def ei(self, x):

        """ Expected Improvement """
        EI=[]
        for o in range(self.problem.DO):
            optimize = self.problem.OS[o].optimize
            # YC: If PSO is given by user, for no-optimize objectives we can simply ignore the objectives' outputs in EI product
            if (self.options['search_algo']=='pso' and optimize == False):
                #print ("o: ", self.problem.OS[o].name, "is not optimize")
                continue
            # YC: If NSGA2 is given by user (e.g. more than three objectives tuning and one no-optimize objective), for no-optimize objectives we still need to return some value for running NSGA2. Maybe we can fix this later.
            elif (optimize == False):
                #print ("o: ", self.problem.OS[o].name, "is not optimize")
                EI.append(0)
            else:
                ymin = self.data.O[self.tid][:,o].min()
                (mu, var) = self.models[o].predict(x, tid=self.tid)
                mu = mu[0][0]
                var = max(1e-18, var[0][0])
                std = np.sqrt(var)
                chi = (ymin - mu) / std
                Phi = 0.5 * (1.0 + sp.special.erf(chi / np.sqrt(2)))
                phi = np.exp(-0.5 * chi**2) / np.sqrt(2 * np.pi * var)
                EI.append(-((ymin - mu) * Phi + var * phi))
                # EI.append(mu)

        if(self.options['search_algo']=='pso' or self.options['search_algo']=='cmaes'):
            EI_prod = np.prod(EI)
            return [-1.0*EI_prod if EI_prod>0 else EI_prod]
        else:
            return EI

    def fitness(self, x):   # x is the normalized space
        xi0 = self.problem.PS.inverse_transform(np.array(x, ndmin=2))
        xi=xi0[0]

        if (any(xx==xi for xx in self.POrig)):
            cond = False
        else:
            point0 = self.D
            point2 = {self.problem.IS[k].name: self.IOrig[k] for k in range(self.problem.DI)}
            point  = {self.problem.PS[k].name: xi[k] for k in range(self.problem.DP)}
            point.update(point0)
            point.update(point2)
            # print("point", point)
            cond = self.computer.evaluate_constraints(self.problem, point)

        if (cond):
            xNorm = self.problem.PS.transform(xi0)[0]
            if(self.problem.models is not None):
                if(self.problem.driverabspath is not None):
                    modulename = Path(self.problem.driverabspath).stem  # get the driver name excluding all directories and extensions
                    sys.path.append(self.problem.driverabspath) # add path to sys
                    module = importlib.import_module(modulename) # import driver name as a module
                else:
                    raise Exception('performance models require passing driverabspath to GPTune')
                # modeldata= self.problem.models(point)
                modeldata= module.models(point)
                xNorm = np.hstack((xNorm,modeldata))  # YL: here tmpdata in the normalized space, but modeldata is the in the original space
                # print(xNorm)

            # print("cond",cond,- self.ei(x),'x',x,'xi',xi)
            #print ("EI: ", self.ei(xNorm))
            return self.ei(xNorm)
        else:
            # print("cond",cond,float("Inf"),'x',x,'xi',xi)
            if(self.options['search_algo']=='pso' or self.options['search_algo']=='cmaes'): # single objective optimizer
                return [float("Inf")]
            else: 
                return [float("Inf")]* self.problem.DO

import pygmo as pg

class SearchPyGMO(Search):

    """
    XXX: This class, together with the underlying PyGMO only works on Intel and AMD CPUs.
    The reason is that PyGMO requires the Intel 'Thread Building Block' library to compile and execute.
    """
    # YL: TBB works also on AMD processors

    def search(self, data : Data, models : Collection[Model], tid : int, **kwargs) -> np.ndarray:

        kwargs = kwargs['kwargs']

        prob = SurrogateProblem(self.problem, self.computer, data, models, self.options, tid)

        try:
            udi = eval(f'pg.{kwargs["search_udi"]}()')
        except:
            raise Exception('Unknown user-defined-island "{kwargs["search_udi"]}"')

        if(kwargs["search_algo"]=='pso' or kwargs["search_algo"]=='cmaes'): # single objective optimizer
            try:
                algo = eval(f'pg.{kwargs["search_algo"]}(gen = kwargs["search_gen"])')
            except:
                raise Exception(f'Unknown optimization algorithm "{kwargs["search_algo"]}"')
            bestX = []
            cond = False
            cpt = 0
            while (not cond and cpt < kwargs['search_max_iters']):
                archi = pg.archipelago(n = kwargs['search_threads'], prob = prob, algo = algo, udi = udi, pop_size = kwargs['search_pop_size'])
                archi.evolve(n = kwargs['search_evolve'])
                archi.wait()
                champions_f = archi.get_champions_f()
                champions_x = archi.get_champions_x()
                indexes = list(range(len(champions_f)))
                indexes.sort(key=champions_f.__getitem__)
                for idx in indexes:
                    if (champions_f[idx] < float('Inf')):
                        cond = True
                        # bestX.append(np.array(self.problem.PS.inverse_transform(np.array(champions_x[idx], ndmin=2))[0]).reshape(1, self.problem.DP))
                        bestX.append(np.array(champions_x[idx]).reshape(1, self.problem.DP))
                        break
                cpt += 1
        else:                   # multi objective
            try:
                algo = eval(f'pg.algorithm(pg.{kwargs["search_algo"]}(gen = kwargs["search_gen"]))')
            except:
                raise Exception(f'Unknown optimization algorithm "{kwargs["search_algo"]}"')
            bestX = []
            cond = False
            cpt = 0
            while (not cond and cpt < kwargs['search_max_iters']):
                pop = pg.population(prob = prob, size = kwargs['search_pop_size'], seed = cpt+1)
                pop = algo.evolve(pop)


                """ It seems pop.get_f() is already sorted, no need to perform the following sorting """
                # if(self.problem.DO==2):
                #   front = pg.non_dominated_front_2d(pop.get_f())
                # else:
                #   ndf, dl, dc, ndr = pg.fast_non_dominated_sorting(pop.get_f())
                #   front = ndf[0]
                # fs = pop.get_f()[front]
                # xs = pop.get_x()[front]
                # bestidx = pg.select_best_N_mo(points = fs, N = kwargs['search_more_samples'])
                # xss = xs[bestidx]
                # fss = fs[bestidx]
                # # print('bestidx',bestidx)

                firstn = min(int(kwargs['search_more_samples']),np.shape(pop.get_f())[0])
                fss = pop.get_f()[0:firstn]
                xss = pop.get_x()[0:firstn]
                # print('firstn',firstn,int(kwargs['search_more_samples']),np.shape(pop.get_f()),xss)


                if(np.max(fss)< float('Inf')):
                    cond = True
                    bestX.append(xss)
                    break
                cpt += 1
        if (kwargs['verbose']):
            print(tid, 'OK' if cond else 'KO'); sys.stdout.flush()
        print("bestX",bestX)
        return (tid, bestX)

##### Simple constrained MOO

class SurrogateProblemCMO(object):

    def __init__(self, problem, computer, data, models, options, tid):   # data is in the normalized space, IOrig and POrig are then generated in the original space

        self.problem = problem
        self.computer = computer
        self.data = data
        self.models = models
        self.options = options

        self.tid = tid

        self.D     = self.data.D[tid]
        self.IOrig = self.problem.IS.inverse_transform(np.array(self.data.I[tid], ndmin=2))[0]
        self.POrig = self.problem.PS.inverse_transform(np.array(self.data.P[tid], ndmin=2))

    def get_nobj(self):
        if(self.options['search_algo']=='pso' or self.options['search_algo']=='cmaes'):
            return 1
        else:
            return self.problem.DO

    def get_bounds(self):

        DP = self.problem.DP

        return ([0. for i in range(DP)], [1. for  i in range(DP)])

    # Acquisition function
    def ei(self, x):

        out_of_range = False
        for o in range(self.problem.DO):
            (lower_bound, upper_bound) = self.problem.OS.bounds[o]
            (mu, var) = self.models[o].predict(x, tid=self.tid)
            mu = mu[0][0]
            var = max(1e-18, var[0][0])
            std = np.sqrt(var)

            print ("mu: ", mu, "var: ", var)
            print ("lower_bound: ", lower_bound, "upper_bound: ", upper_bound)

            #if mu+std < lower_bound or mu-std > upper_bound:
            #    out_of_range = True
            if mu < lower_bound or mu > upper_bound:
                out_of_range = True

        if out_of_range == True:
            return [0]
        else:
            ret = 1
            for o in range(self.problem.DO):
                (mu, var) = self.models[o].predict(x, tid=self.tid)
                ret *= abs(1.0/mu[0][0])
                print ("ret: ", ret)
            return [-1.0*ret]

            #""" Expected Improvement """
            #EI=[]
            #for o in range(self.problem.DO):
            #    ymin = self.data.O[self.tid][:,o].min()
            #    (mu, var) = self.models[o].predict(x, tid=self.tid)
            #    mu = mu[0][0]
            #    var = max(1e-18, var[0][0])
            #    std = np.sqrt(var)
            #    chi = (ymin - mu) / std
            #    Phi = 0.5 * (1.0 + sp.special.erf(chi / np.sqrt(2)))
            #    phi = np.exp(-0.5 * chi**2) / np.sqrt(2 * np.pi * var)
            #    EI.append(-((ymin - mu) * Phi + var * phi))

            ##print ("EI: ", EI)

            #if(self.options['search_algo']=='pso' or self.options['search_algo']=='cmaes'):
            #    EI_prod = np.prod(EI)
            #    return [-1.0*EI_prod if EI_prod>0 else EI_prod]
            #else:
            #    return EI

    def fitness(self, x):   # x is the normalized space
        xi0 = self.problem.PS.inverse_transform(np.array(x, ndmin=2))
        xi=xi0[0]

        if (any(xx==xi for xx in self.POrig)):
            cond = False
        else:
            point0 = self.D
            point2 = {self.problem.IS[k].name: self.IOrig[k] for k in range(self.problem.DI)}
            point  = {self.problem.PS[k].name: xi[k] for k in range(self.problem.DP)}
            point.update(point0)
            point.update(point2)
            # print("point", point)
            cond = self.computer.evaluate_constraints(self.problem, point)

        if (cond):
            xNorm = self.problem.PS.transform(xi0)[0]
            if(self.problem.models is not None):
                if(self.problem.driverabspath is not None):
                    modulename = Path(self.problem.driverabspath).stem  # get the driver name excluding all directories and extensions
                    sys.path.append(self.problem.driverabspath) # add path to sys
                    module = importlib.import_module(modulename) # import driver name as a module
                else:
                    raise Exception('performance models require passing driverabspath to GPTune')
                # modeldata= self.problem.models(point)
                modeldata= module.models(point)
                xNorm = np.hstack((xNorm,modeldata))  # YL: here tmpdata in the normalized space, but modeldata is the in the original space
                # print(xNorm)

            # print("cond",cond,- self.ei(x),'x',x,'xi',xi)
            print ("EI: ", self.ei(xNorm))
            return self.ei(xNorm)
        else:
            # print("cond",cond,float("Inf"),'x',x,'xi',xi)
            if(self.options['search_algo']=='pso' or self.options['search_algo']=='cmaes'): # single objective optimizer
                return [float("Inf")]
            else:
                return [float("Inf")]* self.problem.DO

import pygmo as pg

class SearchCMO(Search):

    """
    XXX: This class, together with the underlying PyGMO only works on Intel and AMD CPUs.
    The reason is that PyGMO requires the Intel 'Thread Building Block' library to compile and execute.
    """

    def search(self, data : Data, models : Collection[Model], tid : int, **kwargs) -> np.ndarray:

        print ("SearchByCMO")

        kwargs = kwargs['kwargs']

        prob = SurrogateProblemCMO(self.problem, self.computer, data, models, self.options, tid)

        try:
            udi = eval(f'pg.{kwargs["search_udi"]}()')
        except:
            raise Exception('Unknown user-defined-island "{kwargs["search_udi"]}"')

        if(kwargs["search_algo"]=='pso' or kwargs["search_algo"]=='cmaes'): # single objective optimizer
            try:
                algo = eval(f'pg.{kwargs["search_algo"]}(gen = kwargs["search_gen"])')
            except:
                raise Exception(f'Unknown optimization algorithm "{kwargs["search_algo"]}"')
            bestX = []
            cond = False
            cpt = 0
            while (not cond and cpt < kwargs['search_max_iters']):
                archi = pg.archipelago(n = kwargs['search_threads'], prob = prob, algo = algo, udi = udi, pop_size = kwargs['search_pop_size'])
                archi.evolve(n = kwargs['search_evolve'])
                archi.wait()
                champions_f = archi.get_champions_f()
                champions_x = archi.get_champions_x()
                indexes = list(range(len(champions_f)))
                indexes.sort(key=champions_f.__getitem__)
                for idx in indexes:
                    if (champions_f[idx] < float('Inf')):
                        cond = True
                        # bestX.append(np.array(self.problem.PS.inverse_transform(np.array(champions_x[idx], ndmin=2))[0]).reshape(1, self.problem.DP))
                        bestX.append(np.array(champions_x[idx]).reshape(1, self.problem.DP))
                        break
                cpt += 1
        else:                   # multi objective
            try:
                algo = eval(f'pg.algorithm(pg.{kwargs["search_algo"]}(gen = kwargs["search_gen"]))')
            except:
                raise Exception(f'Unknown optimization algorithm "{kwargs["search_algo"]}"')
            bestX = []
            cond = False
            cpt = 0
            while (not cond and cpt < kwargs['search_max_iters']):
                pop = pg.population(prob = prob, size = kwargs['search_pop_size'], seed = cpt+1)
                pop = algo.evolve(pop)

                firstn = min(int(kwargs['search_more_samples']),np.shape(pop.get_f())[0])
                fss = pop.get_f()[0:firstn]
                xss = pop.get_x()[0:firstn]
                # print('firstn',firstn,int(kwargs['search_more_samples']),np.shape(pop.get_f()),xss)

                if(np.max(fss)< float('Inf')):
                    cond = True
                    bestX.append(xss)
                    break
                cpt += 1
        if (kwargs['verbose']):
            print(tid, 'OK' if cond else 'KO'); sys.stdout.flush()
        # print("bestX",bestX)
        return (tid, bestX)


if __name__ == '__main__':

    def objectives(point):
        print('this is a dummy definition')
        return point

    def models(point):
        print('this is a dummy definition')
        return point
    
    def models_update(data):
        print('this is a dummy definition')
        return data

    def cst1(point):
        print('this is a dummy definition')
        return point
    def cst2(point):
        print('this is a dummy definition')
        return point
    def cst3(point):
        print('this is a dummy definition')
        return point
    def cst4(point):
        print('this is a dummy definition')
        return point
    def cst5(point):
        print('this is a dummy definition')
        return point
    def cst6(point):
        print('this is a dummy definition')
        return point
    from mpi4py import MPI
    mpi_comm = mpi4py.MPI.Comm.Get_parent()
    mpi_rank = mpi_comm.Get_rank()
    mpi_size = mpi_comm.Get_size()
    (searcher, data, models, tids, kwargs) = mpi_comm.bcast(None, root=0)
    tids_loc = tids[mpi_rank:len(tids):mpi_size]
    tmpdata = searcher.search_multitask(data, models, tids_loc, i_am_manager = False, **kwargs)
    res = mpi_comm.gather(tmpdata, root=0)
    mpi_comm.Disconnect()





# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#

import math

class Options(dict):

    def __init__(self, **kwargs):

        """ Options for GPTune """
        RCI_mode = False         # whether the reverse communication mode will be used
        mpi_comm = None          # The mpi communicator that invokes gptune if mpi4py is installed
        distributed_memory_parallelism = False   # Using distributed_memory_parallelism for the modeling (one MPI per model restart) and search phase (one MPI per task)
        shared_memory_parallelism      = False   # Using shared_memory_parallelism for the modeling (one MPI per model restart) and search phase (one MPI per task)
        constraints_evaluation_parallelism = False  # Reserved option
        verbose = False     # Control the verbosity level
        oversubscribe = False     # Set this to True when the physical core count is less than computer.nodes*computer.cores and the --oversubscribe MPI runtime option is used


        """ Options for the function evaluation """
        objective_evaluation_parallelism   = False  # Using distributed_memory_parallelism or shared_memory_parallelism for evaluating multiple application instances in parallel
        objective_multisample_processes = None  # Number of MPIs each handling one application call
        objective_multisample_threads = None  # Number of threads each handling one application call
        objective_nprocmax = None # Maximum number of cores for each application call, default to computer.cores*computer.nodes-1

        """ Options for the sampling phase """
        sample_class = 'SampleLHSMDU' # Supported sample classes: 'SampleLHSMDU', 'SampleOpenTURNS'
        sample_algo = 'LHS-MDU' # Supported sample algorithms in 'SampleLHSMDU': 'LHS-MDU' --Latin hypercube sampling with multidimensional uniformity, 'MCS' --Monte Carlo Sampling
        sample_max_iter = 10**9  # Maximum number of iterations for generating random samples and testing the constraints



        """ Options for the modeling phase """
        model_class = 'Model_LCM' # Supported sample algorithms: 'Model_GPy_LCM' -- LCM from GPy, 'Model_LCM' -- LCM with fast and parallel inversion, 'Model_DGP' -- deep Gaussian process
        model_output_constraint = False # True: if Model_LCM is used, check output range constraint and disregard out-of-range output (put a large value)
        model_threads = None  # Number of threads used for building one GP model in Model_LCM
        model_processes = None # Number of MPIs used for building one GP model in Model_LCM
        model_groups = 1  # Reserved option
        model_restarts = 1 # Number of random starts each building one initial GP model
        model_restart_processes = None  # Number of MPIs each handling one random start
        model_restart_threads = None   # Number of threads each handling one random start
        model_max_iters = 15000   # Number of maximum iterations for the optimizers
        model_jitter = 1e-10   # Initial jittering
        model_latent = None # Number of latent functions for building one LCM model, defaults to number of tasks
        model_sparse = False # Whether to use SparseGPRegression or SparseGPCoregionalizedRegression from Model_GPy_LCM
        model_inducing = None # Number of inducing points for SparseGPRegression or SparseGPCoregionalizedRegression
        model_layers = 2 # Number of layers for Model_DGP
        model_max_jitter_try = 10 # Max number of jittering 


        """ Options for the search phase """
        search_class = 'SearchPyGMO' #'SearchCMO' #'SearchPyGMO' # Supported searcher classes: 'SearchPyGMO'
        search_threads = None  # Number of threads in each thread group handling one task
        search_processes = 1  # Reserved option
        search_multitask_threads = None # Number of threads groups each handling one task
        search_multitask_processes = None # Number of MPIs each handling one task
        search_algo = 'pso' # Supported search algorithm in pygmo: single-objective: 'pso' -- particle swarm, 'cmaes' -- covariance matrix adaptation evolution. multi-objective 'nsga2' -- Non-dominated Sorting GA, 'nspso' -- Non-dominated Sorting PSO, 'maco' -- Multi-objective Hypervolume-based ACO, 'moead' -- Multi-objective EA vith Decomposition
        search_udi = 'thread_island' # Supported UDI options for pgymo: 'thread_island' --Thread island, 'ipyparallel_island' --Ipyparallel island
        search_pop_size = 1000 # Population size in pgymo
        search_gen = 1000  # Number of evolution generations in pgymo
        search_evolve = 10  # Number of times migration in pgymo
        search_max_iters = 10  # Max number of searches to get results respecting the constraints
        search_more_samples = 1  # Maximum number of points selected using a multi-objective search algorithm


        """ Options for the multi-arm bandit algorithm """
        budget_min = 0.1 # minimum budget
        budget_max = 1 # maximum budget
        budget_base = 2 # the number of arms is floor{log_{budget_base}{budget_max/budget_min}}+1
        fidelity_map = None

        self.update(locals())
        self.update(kwargs)
        self.pop('self')

        self.update(locals())
        self.update(kwargs)
        self.pop('self')


    def validate(self, computer, **kwargs):

        """  modify the options as needed """
        if ((self['model_class']=='Model_LCM' or self['model_class']=='Model_LCM_constrained') and self['RCI_mode']==True):
            self['model_class']='Model_GPy_LCM'

        if (self['distributed_memory_parallelism'] and self['shared_memory_parallelism']):
            self['shared_memory_parallelism']=False

        if (self['distributed_memory_parallelism']):
            if(self['search_multitask_processes'] is None):
                self['search_multitask_processes'] = max(1,computer.cores*computer.nodes-1) # computer.nodes
            self['search_multitask_threads'] = 1
            if(self['model_restart_processes'] is None):
                self['model_restart_processes'] = self['model_restarts']
            self['model_restart_processes'] = min(self['model_restart_processes'],max(1,computer.cores*computer.nodes-2))
            self['model_restart_threads'] = 1
        elif(self['shared_memory_parallelism']):
            self['search_multitask_processes'] = 1
            if(self['search_multitask_threads'] is None):
                self['search_multitask_threads'] = computer.cores
            self['model_restart_processes'] = 1
            if(self['model_restart_threads'] is None):
                self['model_restart_threads'] = self['model_restarts']
            self['model_restart_threads'] = min(self['model_restart_threads'],computer.cores)
        else:
            self['search_multitask_processes'] = 1
            self['search_multitask_threads'] = 1
            self['model_restart_processes'] = 1
            self['model_restart_threads'] = 1

        if (self['model_class']=='Model_LCM' or self['model_class']=='Model_LCM_constrained'):
            if(self['model_processes'] is None):
                if (self['distributed_memory_parallelism']):
                    self['model_processes'] = max(1,math.floor(((computer.cores*computer.nodes-1)/(self['model_restart_processes'])-1)/self['model_restart_threads']))
                else:
                    self['model_processes'] = max(1,math.floor((computer.cores*computer.nodes-1)/self['model_restart_processes']/self['model_restart_threads']))
            self['model_threads'] =1
        else:
            self['model_processes'] = 1
            self['model_threads'] = 1  # YL: this requires more thoughts, maybe math.floor((computer.cores)/self['model_restart_threads'])

        if(self['search_threads'] is None):
            # if (self['distributed_memory_parallelism']):
            #   self['search_threads'] = min(computer.cores,max(1,math.floor((computer.cores*computer.nodes-1)/self['search_multitask_processes'])))
            # else:
            #   self['search_threads'] = min(computer.cores,max(1,math.floor((computer.cores*computer.nodes)/self['search_multitask_threads'])))
            self['search_threads'] =1

        if(self['objective_nprocmax'] is None):
            self['objective_nprocmax'] = computer.cores*computer.nodes-1
        self['objective_nprocmax'] = min(self['objective_nprocmax'],computer.cores*computer.nodes-1)


        if (self['objective_evaluation_parallelism']==True and self['distributed_memory_parallelism']==True):
            self['objective_nprocmax'] = max(1,min(self['objective_nprocmax'],computer.cores*computer.nodes-2))
            nproc = max(1,math.floor((computer.cores*computer.nodes-1)/(self['objective_nprocmax']+1)))  # here we always assume the user invoke application code with MPI_Spawn, if not, "+1" can be removed
            if(self['objective_multisample_processes'] is None):
                self['objective_multisample_processes'] = nproc
            self['objective_multisample_processes'] = min(self['objective_multisample_processes'],nproc)
            self['objective_multisample_threads'] = 1
        elif (self['objective_evaluation_parallelism']==True and self['shared_memory_parallelism']==True):
            nproc = max(1,math.floor((computer.cores*computer.nodes)/(self['objective_nprocmax']+1)))
            if(self['objective_multisample_threads'] is None):
                self['objective_multisample_threads'] = computer.cores
            self['objective_multisample_threads'] = min(self['objective_multisample_threads'],computer.cores,nproc)
            self['objective_multisample_processes'] = 1
        else:
            self['objective_multisample_threads'] = 1
            self['objective_multisample_processes'] = 1

        if(self['verbose']==True):
            print('\n\n------Validating the options')
            print("  ")
            print("  total core counts provided to GPTune:", computer.cores*computer.nodes)
            print("   ---> distributed_memory_parallelism:", self['distributed_memory_parallelism'])
            print("   ---> shared_memory_parallelism:", self['shared_memory_parallelism'])
            print("   ---> objective_evaluation_parallelism:", self['objective_evaluation_parallelism'])


        if(self['distributed_memory_parallelism']):
            ncore_model = (self['model_processes']+1)*self['model_threads']*self['model_restart_processes']+1
        else:
            ncore_model = (self['model_processes']+1)*self['model_threads']*self['model_restart_threads']

        if(self['verbose']==True):    
            print("  ")
            print("  total core counts for modeling:", ncore_model)
            print("   ---> model_processes:", self['model_processes'])
            print("   ---> model_threads:", self['model_threads'])
            print("   ---> model_restart_processes:", self['model_restart_processes'])
            print("   ---> model_restart_threads:", self['model_restart_threads'])


        if(self['distributed_memory_parallelism']):
            ncore_search = self['search_threads']*self['search_multitask_processes']+1
        else:
            ncore_search = self['search_threads']*(self['search_multitask_threads'])

        if(self['verbose']==True):        
            print("  ")
            print("  total core counts for search:", ncore_search)
            print("   ---> search_processes:", self['search_processes'])
            print("   ---> search_threads:", self['search_threads'])
            print("   ---> search_multitask_processes:", self['search_multitask_processes'])
            print("   ---> search_multitask_threads:", self['search_multitask_threads'])

        if(self['distributed_memory_parallelism']):
            if(self['objective_evaluation_parallelism']==True):
                ncore_obj = self['objective_multisample_processes']*(self['objective_nprocmax']+1)+1
            else:
                ncore_obj = (self['objective_nprocmax']+1)
        else:
            ncore_obj = self['objective_multisample_threads']*(self['objective_nprocmax']+1)
        if(self['verbose']==True):        
            print("  ")
            print("  total core counts for objective function evaluation:", ncore_obj)
            print("   ---> core counts in a single application run:", self['objective_nprocmax'])
            print("   ---> objective_multisample_processes:", self['objective_multisample_processes'])
            print("   ---> objective_multisample_threads:", self['objective_multisample_threads'])

        if(self['oversubscribe']==False):
            if (computer.cores*computer.nodes<=1):
                raise Exception("the computer should has at least 2 total cores")

            if ((computer.cores*computer.nodes)<ncore_model):
                raise Exception("Reduce one of the options: model_restart_processes,model_restart_threads,model_processes,model_threads")
            if ((computer.cores*computer.nodes)<ncore_search):
                raise Exception("Reduce one of the options: search_multitask_processes,search_multitask_threads,search_processes,search_threads")
            if ((computer.cores*computer.nodes)<ncore_obj):
                raise Exception("Reduce one of the options: objective_multisample_processes,objective_multisample_threads,objective_nprocmax")




# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#

import abc
from typing import Callable
import numpy as np
import math
import skopt.space
from skopt.space import *
import time

from autotune.space import Space

class Sample(abc.ABC):

    @abc.abstractmethod
    def sample(self, n_samples : int, space : Space, **kwargs):

        raise Exception("Abstract method")

    def sample_constrained(self, n_samples : int, space : Space, check_constraints : Callable = None, check_constraints_kwargs : dict = {}, **kwargs):

        if (check_constraints is None):
            S = self.sample(n_samples, space)

        else:

            if ('sample_max_iter' in kwargs):
                sample_max_iter = kwargs['sample_max_iter']
            else:
                if ('options' in kwargs):
                    sample_max_iter = kwargs['options']['sample_max_iter']
                else:
                    sample_max_iter = 1

            S = []
            cpt = 0
            n_itr = 0
            while ((cpt < n_samples) and (n_itr < sample_max_iter)):
                # t1 = time.time_ns()
                S2 = self.sample(n_samples, space, kwargs=kwargs)
                # t2 = time.time_ns()
                # print('sample_para:',(t2-t1)/1e9)

                for s_norm in S2:
                    # print("jiji",s_norm)
                    s_orig = space.inverse_transform(np.array(s_norm, ndmin=2))[0]
                    kwargs2 = {d.name: s_orig[i] for (i, d) in enumerate(space)}
                    # print("dfdfdfdfd",kwargs2)
                    kwargs2.update(check_constraints_kwargs)
                    if (check_constraints(kwargs2)):
                        S.append(s_norm)
                        cpt += 1
                        if (cpt >= n_samples):
                            break
                # print('input',S,space[0],isinstance(space[0], Categorical))

                n_itr += 1
                if(n_itr%1000==0 and n_itr>=1000):
                    print('n_itr',n_itr,'still trying generating constrained samples...')


            if (cpt < n_samples):
                raise Exception("Only %d valid samples were generated while %d were requested.\
                        The constraints might be too hard to satisfy.\
                        Consider increasing 'sample_max_iter', or, provide a user-defined sampling method."%(len(S), n_samples))
        # print('reqi',S,'nsample',n_samples,sample_max_iter,space)
        S = np.array(S[0:n_samples]).reshape((n_samples, len(space)))

        return S

    def sample_inputs(self, n_samples : int, IS : Space, check_constraints : Callable = None, check_constraints_kwargs : dict = {}, **kwargs):

        return self.sample_constrained(n_samples, IS, check_constraints = check_constraints, check_constraints_kwargs = check_constraints_kwargs, **kwargs)

    def sample_parameters(self, n_samples : int, I : np.ndarray, IS : Space, PS : Space, check_constraints : Callable = None, check_constraints_kwargs : dict = {}, **kwargs):



        P = []
        for t in I:
            # print('before inverse_transform:',np.array(t, ndmin=2))
            I_orig = IS.inverse_transform(np.array(t, ndmin=2))[0]
            # I_orig = t
            # print('after inverse_transform I_orig:',I_orig)
            kwargs2 = {d.name: I_orig[i] for (i, d) in enumerate(IS)}
            kwargs2.update(check_constraints_kwargs)
            xs = self.sample_constrained(n_samples, PS, check_constraints = check_constraints, check_constraints_kwargs = kwargs2, **kwargs)
            P.append(xs)



        return P


import lhsmdu

class SampleLHSMDU(Sample):

    def __init__(self):

        super().__init__()

        self.cached_n_samples = None
        self.cached_space     = None
        self.cached_algo      = None

    def sample(self, n_samples : int, space : Space, **kwargs):

        kwargs = kwargs['kwargs']

        if (self.cached_n_samples is not None and self.cached_n_samples == n_samples and self.cached_space is not None and space == self.cached_space and self.cached_algo is not None and self.cached_algo == kwargs['sample_algo']):

            #lhs = lhsmdu.resample() # YC: this can fall into too clustered samples if there are many constraints
            lhs = lhsmdu.sample(len(space), n_samples)

        else:

            self.cached_n_samples = n_samples
            self.cached_space     = space
            self.cached_algo      = kwargs['sample_algo']

            if (kwargs['sample_algo'] == 'LHS-MDU'):
                lhs = lhsmdu.sample(len(space), n_samples)
            elif (kwargs['sample_algo'] == 'MCS'):
                lhs = lhsmdu.createRandomStandardUniformMatrix(len(space), n_samples)
            else:
                raise Excepetion(f"Unknown algorithm {kwargs['sample_algo']}")

        lhs = np.array(list(zip(*[np.array(lhs[k])[0] for k in range(len(lhs))])))
        # print(lhs,'normalized',n_samples)

        return lhs


import openturns as ot

class SampleOpenTURNS(Sample):

    """
    XXX: This class, together with the underlying OpenTURNS only works on Intel and AMD CPUs.
    The reason is that OpenTURNS requires the Intel 'Thread Building Block' library to compile and execute.
    """

    def __init__(self):

        super().__init__()

        self.cached_space        = None
        self.cached_distribution = None

    def sample(self, n_samples : int, space : Space, **kwargs):

        if (self.cached_space is not None and space == self.cached_space):
            distribution = self.cached_distribution
        else:
            distributions = [ot.Uniform(*d.transformed_bounds) for d in space.dimensions]
            distribution  = ot.ComposedDistribution(distributions)
            # Caching space and distribution to speed-up future samplings, especially if invoked by the sample_constrained method.
            self.cached_space = space
            self.cached_distribution = distribution

        lhs = ot.LHSExperiment(distribution, n_samples)
        lhs.setAlwaysShuffle(True) # randomized

        S = lhs.generate()
        S = np.array(S)
        S.reshape((n_samples, len(space)))

        return S




# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#




# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#

import numpy as np
from problem import Problem
from data import Data
import json
import os.path
from filelock import FileLock
from autotune.space import *
from autotune.problem import TuningProblem
import uuid
import time
import requests

def GetMachineConfiguration(meta_path=None, meta_dict=None):
    import ast

    # machine configuration values
    machine_name = "mymachine"
    processor_model = "unknown"
    nodes = 1
    cores = 2

    if (os.environ.get('CKGPTUNE_HISTORY_DB') == 'yes'):
        try:
            machine_configuration = ast.literal_eval(os.environ.get('CKGPTUNE_MACHINE_CONFIGURATION', '{}'))
            machine_name = machine_configuration['machine_name']
            processor_list = list(machine_configuration.keys())
            processor_list.remove('machine_name')
            # YC: we currently assume the application uses only one processor type
            processor_model = processor_list[0]
            nodes = machine_configuration[processor_model]['nodes']
            cores = machine_configuration[processor_model]['cores']
        except:
            print ("[HistoryDB] not able to get machine configuration")

    else:
        if meta_path != None or meta_dict != None:
            metadata = {}

            if meta_path != None and os.path.exists(meta_path):
                try:
                    with open(meta_path) as f_in:
                        metadata.update(json.load(f_in))
                except:
                    print ("[Error] not able to load meta description from path")

            if meta_dict != None:
                try:
                    metadata.update(meta_dict)
                except:
                    print ("[Error] not able to load meta description from dict")

            machine_configuration = metadata['machine_configuration']
            machine_name = machine_configuration['machine_name']
            processor_list = list(machine_configuration.keys())
            processor_list.remove('machine_name')
            # YC: we currently assume the application uses only one processor type
            processor_model = processor_list[0]
            nodes = machine_configuration[processor_model]['nodes']
            cores = machine_configuration[processor_model]['cores']

        elif os.path.exists("./.gptune/meta.json"):
            try:
                with open("./.gptune/meta.json") as f_in:
                    gptune_metadata = json.load(f_in)
                    machine_configuration = gptune_metadata['machine_configuration']
                    machine_name = machine_configuration['machine_name']
                    processor_list = list(machine_configuration.keys())
                    processor_list.remove('machine_name')
                    # YC: we currently assume the application uses only one processor type
                    processor_model = processor_list[0]
                    nodes = machine_configuration[processor_model]['nodes']
                    cores = machine_configuration[processor_model]['cores']
            except:
                print ("[HistoryDB] not able to get machine configuration")

        else:
            print ("[HistoryDB] not able to get machine configuration")

    #return (machine_configuration)
    return (machine_name, processor_model, nodes, cores)

def GetMachineConfigurationDict(meta_description_path = "./.gptune/meta.json"):
    import ast

    machine_configuration = {}

    if (os.environ.get('CKGPTUNE_HISTORY_DB') == 'yes'):
        try:
            machine_configuration = ast.literal_eval(os.environ.get('CKGPTUNE_MACHINE_CONFIGURATION', '{}'))
        except:
            print ("[HistoryDB] not able to get machine configuration")

    elif os.path.exists(meta_description_path):
        try:
            with open(meta_description_path) as f_in:
                gptune_metadata = json.load(f_in)
                machine_configuration = gptune_metadata['machine_configuration']
        except:
            print ("[HistoryDB] not able to get machine configuration")

    else:
        print ("[HistoryDB] not able to get machine configuration")

    return (machine_configuration)

def GetSoftwareConfigurationDict(meta_description_path = "./.gptune/meta.json"):
    import ast

    software_configuration = {}

    if (os.environ.get('CKGPTUNE_HISTORY_DB') == 'yes'):
        try:
            software_configuration = ast.literal_eval(os.environ.get('CKGPTUNE_SOFTWARE_CONFIGURATION', '{}'))
        except:
            print ("[HistoryDB] not able to get software configuration")

    elif os.path.exists(meta_description_path):
        try:
            with open(meta_description_path) as f_in:
                gptune_metadata = json.load(f_in)
                software_configuration = gptune_metadata['software_configuration']
        except:
            print ("[HistoryDB] not able to get software configuration")

    else:
        print ("[HistoryDB] not able to get software configuration")

    return (software_configuration)

def search_item_by_uid(dict_arr, uid):

    for item in dict_arr:
        if (item["uid"] == uid):
            return item

    return None

class HistoryDB(dict):

    def __init__(self, meta_path=None, meta_dict=None, **kwargs):

        self.tuning_problem_name = None
        self.tuning_problem_category = None

        """ Options """
        self.history_db = True
        self.save_func_eval = True
        self.save_model = True
        self.load_func_eval = True
        self.load_surrogate_model = False

        """ Crowd repository options """
        self.use_crowd_repo = False
        self.historydb_api_key = ""
        self.crowd_repo_download_url = "https://gptune.lbl.gov/repo/direct-download/" # GPTune HistoryDB repo
        #self.crowd_repo_download_url = "http://127.0.0.1:8000/repo/direct-download/" # debug
        self.crowd_repo_upload_url = "https://gptune.lbl.gov/repo/direct-upload/" # GPTune HistoryDB repo
        #self.crowd_repo_upload_url = "http://127.0.0.1:8000/repo/direct-upload/" # debug

        """ Path to JSON data files """
        self.history_db_path = "./"

        """ Pass machine-related information """
        self.machine_configuration = {
                    "machine":"Unknown",
                    "nodes":"Unknown",
                    "cores":"Unknown"
                }

        """ Pass software-related information """
        self.software_configuration = {}

        """ Loadable machine configurations """
        self.loadable_machine_configurations = {}

        """ Loadable software configurations """
        self.loadable_software_configurations = {}

        """ Verbose debug message """
        self.verbose = 1

        """ list of UIDs of function evaluation results """
        self.uids = []

        """ File synchronization options """
        self.file_synchronization_method = 'filelock'

        """ Process uid """
        self.process_uid = str(uuid.uuid1())

        # if history database is requested by CK-GPTune
        if (os.environ.get('CKGPTUNE_HISTORY_DB') == 'yes'):
            print ("CK-GPTune History Database Init")
            import ast
            self.tuning_problem_name = os.environ.get('CKGPTUNE_TUNING_PROBLEM_NAME','Unknown')
            self.machine_configuration = ast.literal_eval(os.environ.get('CKGPTUNE_MACHINE_CONFIGURATION','{}'))

            software_configuration = ast.literal_eval(os.environ.get('CKGPTUNE_SOFTWARE_CONFIGURATION','{}'))
            for x in software_configuration:
                self.software_configuration[x] = software_configuration[x]
            compile_deps = ast.literal_eval(os.environ.get('CKGPTUNE_COMPILE_DEPS','{}'))
            for x in compile_deps:
                self.software_configuration[x] = compile_deps[x]
            runtime_deps = ast.literal_eval(os.environ.get('CKGPTUNE_RUNTIME_DEPS','{}'))
            for x in runtime_deps:
                self.software_configuration[x] = runtime_deps[x]

            self.loadable_machine_configurations = ast.literal_eval(os.environ.get('CKGPTUNE_LOADABLE_MACHINE_CONFIGURATIONS','{}'))
            self.loadable_software_configurations = ast.literal_eval(os.environ.get('CKGPTUNE_LOADABLE_SOFTWARE_CONFIGURATIONS', '{}'))

            os.system("mkdir -p ./gptune.db")
            self.history_db_path = "./gptune.db"

            if (os.environ.get('CKGPTUNE_LOAD_MODEL') == 'yes'):
                self.load_surrogate_model = True
            try:
                with FileLock("test.lock", timeout=0):
                    print ("[HistoryDB] use filelock for synchronization")
                    self.file_synchronization_method = 'filelock'
            except:
                print ("[HistoryDB] use rsync for synchronization")
                self.file_synchronization_method = 'rsync'
            os.system("rm -rf test.lock")

        # if GPTune is called through MPI spawning or Reverse Communication Interface
        else:
            metadata = {}

            if meta_path != None or meta_dict != None:
                if meta_path != None and os.path.exists(meta_path):
                    try:
                        with open(meta_path) as f_in:
                            metadata.update(json.load(f_in))
                    except:
                        print ("[Error] not able to load meta description from path")
                if meta_dict != None:
                    try:
                        metadata.update(meta_dict)
                    except:
                        print ("[Error] not able to load meta description from dict")

            elif os.path.exists('./.gptune/meta.json'): #or (os.environ.get('GPTUNE_RCI') == 'yes'):
                with open("./.gptune/meta.json") as f_in:
                    print ("GPTune History Database Init")

                    metadata = json.load(f_in)
            else:
                self.history_db = False
                raise Exception("History database initialization failed")

            if "tuning_problem_name" in metadata:
                self.tuning_problem_name = metadata["tuning_problem_name"]
            else:
                self.tuning_problem_name = "Unknown"

            if "tuning_problem_category" in metadata:
                self.tuning_problem_category = metadata["tuning_problem_category"]
            else:
                self.tuning_problem_category = "Unknown"

            if "use_crowd_repo" in metadata:
                if metadata["use_crowd_repo"] == "yes" or metadata["use_crowd_repo"] == "y":
                    self.use_crowd_repo = True
                else:
                    self.use_crowd_repo = False
            else:
                self.use_crowd_repo = False

            if "historydb_api_key" in metadata:
                self.historydb_api_key = metadata["historydb_api_key"]
            else:
                self.historydb_api_key = ""

            if "history_db_path" in metadata:
                self.history_db_path = metadata["history_db_path"]
            else:
                os.system("mkdir -p ./gptune.db")
                self.history_db_path = "./gptune.db"

            if "machine_configuration" in metadata:
                self.machine_configuration = metadata["machine_configuration"]
            if "software_configuration" in metadata:
                self.software_configuration = metadata["software_configuration"]
            if "loadable_machine_configurations" in metadata:
                self.loadable_machine_configurations = metadata["loadable_machine_configurations"]
            if "loadable_software_configurations" in metadata:
                self.loadable_software_configurations = metadata["loadable_software_configurations"]
            if "spack" in metadata:
                spack_loaded_items = metadata["spack"]
                for spack_loaded_item in spack_loaded_items:
                    import subprocess

                    stdout, stderr = subprocess.Popen("spack find --json "+spack_loaded_item, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()

                    try:
                        json_data = json.loads(stdout)[0]

                        software_name = json_data["name"]
                        version_split = [int(v) for v in json_data["version"].split(".")]

                        self.software_configuration[software_name] = { "version_split" : version_split }

                        for software_depend in json_data["dependencies"]:
                            software_depend_json = json_data["dependencies"][software_depend]
                            hash_value = software_depend_json["hash"]
                            types = software_depend_json["type"]
                            if "build" in types and "link" in types:
                                stdout, stderr = subprocess.Popen("spack find --json "+software_depend, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
                                software_depend_spec_found = json.loads(stdout)
                                for software_depend_spec in software_depend_spec_found:
                                    if software_depend_spec["hash"] == hash_value:
                                        software_depend_name = software_depend_spec["name"]
                                        version_split = [int(v) for v in software_depend_spec["version"].split(".")]
                                        self.software_configuration[software_depend_name] = { "version_split" : version_split }
                                        break
                    except:
                        print ("spack find failed: ", spack_loaded_item)

                    print (self.software_configuration)
            try:
                with FileLock("test.lock", timeout=0):
                    print ("[HistoryDB] use filelock for synchronization")
                    self.file_synchronization_method = 'filelock'
            except:
                print ("[HistoryDB] use rsync for synchronization")
                self.file_synchronization_method = 'rsync'
            os.system("rm -rf test.lock")

    def check_load_deps(self, func_eval):

        ''' check machine configuration dependencies '''
        loadable_machine_configurations = self.loadable_machine_configurations
        machine_configuration = func_eval['machine_configuration']
        machine_name = machine_configuration['machine_name']
        processor_list = list(machine_configuration.keys())
        processor_list.remove("machine_name")
        if not machine_name in loadable_machine_configurations.keys():
            print (machine_name+": " + machine_name + " is not in load_deps: " + str(loadable_machine_configurations.keys()))
            return False
        else:
            for processor in processor_list:
                if not processor in loadable_machine_configurations[machine_name]:
                    return False
                else:
                    num_nodes = machine_configuration[processor]["nodes"]
                    num_nodes_loadable = loadable_machine_configurations[machine_name][processor]["nodes"]
                    if type(num_nodes_loadable) == list:
                        if num_nodes not in num_nodes_loadable:
                            return False
                    elif type(num_nodes_loadable) == int:
                        if num_nodes != num_nodes_loadable:
                            return False
                    else:
                        return False

                    num_cores = machine_configuration[processor]["cores"]
                    num_cores_loadable = loadable_machine_configurations[machine_name][processor]["cores"]
                    if type(num_cores_loadable) == list:
                        if num_cores not in num_cores_loadable:
                            return False
                    elif type(num_cores_loadable) == int:
                        if num_cores != num_cores_loadable:
                            return False
                    else:
                        return False

        ''' check compile-level software dependencies '''
        loadable_software_configurations = self.loadable_software_configurations
        software_configuration = func_eval['software_configuration']
        for software_name in loadable_software_configurations.keys():
            deps_passed = False

            #software_name = loadable_software_configurations[software_name][option]['name']
            if software_name in software_configuration.keys():
                version_split = software_configuration[software_name]['version_split']
                version_value = version_split[0]*100+version_split[1]*10+version_split[2]
                #print ("software_name: " + software_name + " version_value: " + str(version_value))

                if 'version_split' in loadable_software_configurations[software_name].keys():
                    version_dep_split = loadable_software_configurations[software_name]['version_split']
                    version_dep_value = version_dep_split[0]*100+version_dep_split[1]*10+version_dep_split[2]

                    if version_dep_value == version_value:
                        deps_passed = True

                if 'version_from' in loadable_software_configurations[software_name].keys() and \
                   'version_to' not in loadable_software_configurations[software_name].keys():
                    version_dep_from_split = loadable_software_configurations[software_name]['version_from']
                    version_dep_from_value = version_dep_from_split[0]*100+version_dep_from_split[1]*10+version_dep_from_split[2]

                    if version_dep_from_value <= version_value:
                        deps_passed = True

                if 'version_from' not in loadable_software_configurations[software_name].keys() and \
                   'version_to' in loadable_software_configurations[software_name].keys():
                    version_dep_to_split = loadable_software_configurations[software_name]['version_to']
                    version_dep_to_value = version_dep_to_split[0]*100+version_dep_to_split[1]*10+version_dep_to_split[2]

                    if version_dep_to_value >= version_value:
                        deps_passed = True

                if 'version_from' in loadable_software_configurations[software_name].keys() and \
                   'version_to' in loadable_software_configurations[software_name].keys():
                    version_dep_from_split = loadable_software_configurations[software_name]['version_from']
                    version_dep_from_value = version_dep_from_split[0]*100+version_dep_from_split[1]*10+version_dep_from_split[2]

                    version_dep_to_split = loadable_software_configurations[software_name]['version_to']
                    version_dep_to_value = version_dep_to_split[0]*100+version_dep_to_split[1]*10+version_dep_to_split[2]

                    if version_dep_from_value <= version_value and \
                       version_dep_to_value >= version_value:
                        deps_passed = True

            if (deps_passed == False):
                if (self.verbose):
                    print ("deps_passed failed: "  + " " + str(software_name)) 
                return False

        return True

    def search_func_eval_task_id(self, func_eval : dict, problem : Problem, Igiven : np.ndarray):
        task_id = -1

        for i in range(len(Igiven)):
            compare_all_elems = True
            for j in range(len(problem.IS)):
                if (func_eval["task_parameter"][problem.IS[j].name] != Igiven[i][j]):
                    compare_all_elems = False
                    break
            if compare_all_elems == True:
                task_id = i
                break

        return task_id

    def is_parameter_duplication(self, problem : Problem, PS_history, parameter):
        # for i in range(len(PS_history)):
        for j in range(len(PS_history)):
            compare_all_elems = True
            for k in range(len(problem.PS)):
                if (parameter[problem.PS[k].name] != PS_history[j][k]):
                    compare_all_elems = False
                    break
            if compare_all_elems == True:
                print ("found a duplication of parameter set: ", parameter)
                return True

        return False

    def load_history_func_eval(self, data : Data, problem : Problem, Igiven : np.ndarray):
        """ Init history database JSON file """
        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"

            if self.use_crowd_repo == True:
                try:
                    r = requests.get(url = self.crowd_repo_download_url,
                            headers={"x-api-key":self.historydb_api_key},
                            params={"tuning_problem_name":self.tuning_problem_name,
                                "tuning_problem_category":self.tuning_problem_category},
                            verify=False)
                    if r.status_code == 200:
                        if not os.path.exists(json_data_path): #TODO: check
                            with open(json_data_path, "w") as f_out:
                                json_data = {"tuning_problem_name":self.tuning_problem_name,
                                    "tuning_problem_category":self.tuning_problem_category,
                                    "surrogate_model":[],
                                    "func_eval":[]}
                                json.dump(json_data, f_out, indent=2)

                        func_eval_list_downloaded = json.loads(r.text)["perf_data"]
                        print ("FUNC_EVAL_LIST_DOWNLOADED: ", func_eval_list_downloaded)
                        with open(json_data_path, "r") as f_in:
                            json_data = json.load(f_in)
                            json_data["func_eval"] += func_eval_list_downloaded #TODO: uid check
                        with open(json_data_path, "w") as f_out:
                            json.dump(json_data, f_out, indent=2)
                    else:
                        print ("request status_code: ", r.status_code)
                except:
                    print ("direct download failed")

            if os.path.exists(json_data_path):
                print ("[HistoryDB] Found a history database file")
                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "r") as f_in:
                            history_data = json.load(f_in)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        history_data = json.load(f_in)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "r") as f_in:
                        history_data = json.load(f_in)

                num_tasks = len(Igiven)

                num_loaded_data = 0

                PS_history = [[] for i in range(num_tasks)]
                OS_history = [[] for i in range(num_tasks)]

                for func_eval in history_data["func_eval"]:
                    if (self.check_load_deps(func_eval)):
                        task_id = self.search_func_eval_task_id(func_eval, problem, Igiven)
                        if (task_id != -1):
                            # # current policy: skip loading the func eval result
                            # # if the same parameter data has been loaded once (duplicated)
                            # # YL: only need to search in PS_history[task_id], not PS_history
                            # if self.is_parameter_duplication(problem, PS_history[task_id], func_eval["tuning_parameter"]):
                            
                            # current policy: allow duplicated samples 
                            # YL: This makes RCI-based multi-armed bandit much easier to implement, maybe we can add an option for changing this behavior 
                            if False: # self.is_parameter_duplication(problem, PS_history[task_id], func_eval["tuning_parameter"]):
                                continue
                            else:
                                parameter_arr = []
                                for k in range(len(problem.PS)):
                                    if type(problem.PS[k]).__name__ == "Categoricalnorm":
                                        parameter_arr.append(str(func_eval["tuning_parameter"][problem.PS[k].name]))
                                    elif type(problem.PS[k]).__name__ == "Integer":
                                        parameter_arr.append(int(func_eval["tuning_parameter"][problem.PS[k].name]))
                                    elif type(problem.PS[k]).__name__ == "Real":
                                        parameter_arr.append(float(func_eval["tuning_parameter"][problem.PS[k].name]))
                                    else:
                                        parameter_arr.append(func_eval["tuning_parameter"][problem.PS[k].name])
                                PS_history[task_id].append(parameter_arr)
                                OS_history[task_id].append(\
                                    [func_eval["evaluation_result"][problem.OS[k].name] \
                                    for k in range(len(problem.OS))])
                                num_loaded_data += 1

                if (num_loaded_data > 0):
                    data.I = Igiven #IS_history
                    data.P = PS_history
                    data.O=[] # YL: OS is a list of 2D numpy arrays
                    for i in range(len(OS_history)):
                        if(len(OS_history[i])==0):
                            data.O.append(np.empty( shape=(0, problem.DO)))
                        else:
                            data.O.append(np.array(OS_history[i]))
                            if(any(ele==[None] for ele in OS_history[i])):
                                print ("history data contains null function values")
                                exit()
                    # print ("db: data.I: " + str(data.I))
                    # print ("db: data.P: " + str(data.P))
                    # print ("db: data.O: " + str(OS_history))
                else:
                    print ("no history data has been loaded")
            else:
                print ("[HistoryDB] Create a JSON file at " + json_data_path)

                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "w") as f_out:
                            json_data = {"tuning_problem_name":self.tuning_problem_name,
                                "tuning_problem_category":self.tuning_problem_category,
                                "surrogate_model":[],
                                "func_eval":[]}
                            json.dump(json_data, f_out, indent=2)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    with open(temp_path, "w") as f_out:
                        json_data = {"tuning_problem_name":self.tuning_problem_name,
                            "tuning_problem_category":self.tuning_problem_category,
                            "surrogate_model":[],
                            "func_eval":[]}
                        json.dump(json_data, f_out, indent=2)
                    os.system("rsync -u " + temp_path + " " + json_data_path)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "w") as f_out:
                        json_data = {"tuning_problem_name":self.tuning_problem_name,
                            "tuning_problem_category":self.tuning_problem_category,
                            "surrogate_model":[],
                            "func_eval":[]}
                        json.dump(json_data, f_out, indent=2)

    def load_model_func_eval(self, data : Data, problem : Problem, Igiven : np.ndarray, model_data : dict):
        """ Init history database JSON file """
        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"
            if os.path.exists(json_data_path):
                print ("[HistoryDB] Found a history database file")
                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "r") as f_in:
                            history_data = json.load(f_in)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        history_data = json.load(f_in)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "r") as f_in:
                        history_data = json.load(f_in)

                num_tasks = len(Igiven)

                num_loaded_data = 0

                PS_history = [[] for i in range(num_tasks)]
                OS_history = [[] for i in range(num_tasks)]

                # Assume that all function evaluations of the surrogate model are in the database file
                for func_eval_uid in model_data["function_evaluations"]:
                    func_eval = search_item_by_uid(history_data["func_eval"], func_eval_uid)
                    parameter_arr = []
                    for k in range(len(problem.PS)):
                        if type(problem.PS[k]).__name__ == "Categoricalnorm":
                            parameter_arr.append(str(func_eval["tuning_parameter"][problem.PS[k].name]))
                        elif type(problem.PS[k]).__name__ == "Integer":
                            parameter_arr.append(int(func_eval["tuning_parameter"][problem.PS[k].name]))
                        elif type(problem.PS[k]).__name__ == "Real":
                            parameter_arr.append(float(func_eval["tuning_parameter"][problem.PS[k].name]))
                        else:
                            parameter_arr.append(func_eval["tuning_parameter"][problem.PS[k].name])
                    task_id = self.search_func_eval_task_id(func_eval, problem, Igiven)
                    PS_history[task_id].append(parameter_arr)
                    OS_history[task_id].append(\
                        [func_eval["evaluation_result"][problem.OS[k].name] \
                        for k in range(len(problem.OS))])
                    num_loaded_data += 1

                if (num_loaded_data > 0):
                    data.I = Igiven #IS_history
                    data.P = PS_history
                    data.O=[] # YL: OS is a list of 2D numpy arrays
                    for i in range(len(OS_history)):
                        if(len(OS_history[i])==0):
                            data.O.append(np.empty( shape=(0, problem.DO)))
                        else:
                            data.O.append(np.array(OS_history[i]))
                            if(any(ele==[None] for ele in OS_history[i])):
                                print ("history data contains null function values")
                                exit()
                else:
                    print ("no history data has been loaded")
            else:
                print ("[HistoryDB] Create a JSON file at " + json_data_path)

                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "w") as f_out:
                            json_data = {"tuning_problem_name":self.tuning_problem_name,
                                "tuning_problem_category":self.tuning_problem_category,
                                "surrogate_model":[],
                                "func_eval":[]}
                            json.dump(json_data, f_out, indent=2)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    with open(temp_path, "w") as f_out:
                        json_data = {"tuning_problem_name":self.tuning_problem_name,
                            "tuning_problem_category":self.tuning_problem_category,
                            "surrogate_model":[],
                            "func_eval":[]}
                        json.dump(json_data, f_out, indent=2)
                    os.system("rsync -u " + temp_path + " " + json_data_path)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "w") as f_out:
                        json_data = {"tuning_problem_name":self.tuning_problem_name,
                            "tuning_problem_category":self.tuning_problem_category,
                            "surrogate_model":[],
                            "func_eval":[]}
                        json.dump(json_data, f_out, indent=2)

    def store_func_eval(self, problem : Problem,\
            task_parameter : np.ndarray,\
            tuning_parameter : np.ndarray,\
            evaluation_result : np.ndarray,\
            evaluation_detail : np.ndarray,\
            source : str = "measure"):

        print ("store_func_eval")
        print ("problem.constants")
        print (problem.constants)

        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"

            new_function_evaluation_results = []

            now = time.localtime()

            # get the types of each parameter
            task_dtype=''
            for p in problem.IS.dimensions:
                if (isinstance(p, Real)):
                    task_dtype=task_dtype+', float64'
                elif (isinstance(p, Integer)):
                    task_dtype=task_dtype+', int32'
                elif (isinstance(p, Categorical)):
                    task_dtype=task_dtype+', U100'
            task_dtype=task_dtype[2:]
            
            tuning_dtype=''
            for p in problem.PS.dimensions:
                if (isinstance(p, Real)):
                    tuning_dtype=tuning_dtype+', float64'
                elif (isinstance(p, Integer)):
                    tuning_dtype=tuning_dtype+', int32'
                elif (isinstance(p, Categorical)):
                    tuning_dtype=tuning_dtype+', U100'
            tuning_dtype=tuning_dtype[2:]

            # transform to the original parameter space
            task_parameter_orig = problem.IS.inverse_transform(np.array(task_parameter, ndmin=2))[0]
            task_parameter_orig_list = np.array(tuple(task_parameter_orig),dtype=task_dtype).tolist()

            num_evals = len(tuning_parameter)
            for i in range(num_evals):
                uid = uuid.uuid1()
                self.uids.append(str(uid))

                tuning_parameter_orig = problem.PS.inverse_transform(
                        np.array(tuning_parameter[i], ndmin=2))[0]
                tuning_parameter_orig_list = np.array(tuple(tuning_parameter_orig),dtype=tuning_dtype).tolist()
                evaluation_result_orig_list = np.array(evaluation_result[i]).tolist()
                evaluation_result_orig_list = [round(val, 6) for val in evaluation_result_orig_list]
                evaluation_detail_orig_list = np.array(evaluation_detail[i]).tolist()

                task_parameter_store = { problem.IS[k].name:task_parameter_orig_list[k] for k in range(len(problem.IS)) }
                tuning_parameter_store = { problem.PS[k].name:tuning_parameter_orig_list[k] for k in range(len(problem.PS)) }
                constants_store = problem.constants
                if constants_store == None:
                    constants_store = {}
                machine_configuration_store = self.machine_configuration
                software_configuration_store = self.software_configuration
                evaluation_result_store = { problem.OS[k].name:None if np.isnan(evaluation_result_orig_list[k]) else evaluation_result_orig_list[k] for k in range(len(problem.OS)) }
                evaluation_detail_store = { problem.OS[k].name:{} for k in range(len(problem.OS)) }
                for k in range(len(problem.OS)):
                    evaluation_detail_store[problem.OS[k].name] = {}
                    evaluation_detail_store[problem.OS[k].name]["evaluations"] = evaluation_detail_orig_list[k]
                    evaluation_detail_store[problem.OS[k].name]["objective_scheme"] = "average"

                if "machine_configuration" in task_parameter_store:
                    import ast
                    machine_configuration_store = ast.literal_eval(task_parameter_store["machine_configuration"])
                    del task_parameter_store["machine_configuration"]
                if "software_configuration" in task_parameter_store:
                    import ast
                    software_configuration_store = ast.literal_eval(task_parameter_store["software_configuration"])
                    del task_parameter_store["software_configuration"]

                function_evaluation_document = {
                        "task_parameter":task_parameter_store,
                        "tuning_parameter":tuning_parameter_store,
                        "constants":constants_store,
                        "machine_configuration":machine_configuration_store,
                        "software_configuration":software_configuration_store,
                        "evaluation_result":evaluation_result_store,
                        "evaluation_detail":evaluation_detail_store,
                        "source": source,
                        "time":{
                            "tm_year":now.tm_year,
                            "tm_mon":now.tm_mon,
                            "tm_mday":now.tm_mday,
                            "tm_hour":now.tm_hour,
                            "tm_min":now.tm_min,
                            "tm_sec":now.tm_sec,
                            "tm_wday":now.tm_wday,
                            "tm_yday":now.tm_yday,
                            "tm_isdst":now.tm_isdst
                            },
                        "uid":str(uid)
                    }

                new_function_evaluation_results.append(function_evaluation_document)

                if self.use_crowd_repo == True:
                    print ("function_evaluation_document: ", str(function_evaluation_document))
                    print ("API_KEY: ", self.historydb_api_key)

                    try:
                        r = requests.post(url = self.crowd_repo_upload_url,
                                headers={"x-api-key":self.historydb_api_key},
                                data={"tuning_problem_name":self.tuning_problem_name,
                                    "tuning_problem_category":self.tuning_problem_category,
                                    "function_evaluation_document":json.dumps(function_evaluation_document)},
                                verify=False)
                        if r.status_code == 200:
                            print ("direct upload success")
                        else:
                            print ("request status_code: ", r.status_code)
                    except:
                        print ("direct upload failed")

            if self.file_synchronization_method == 'filelock':
                with FileLock(json_data_path+".lock"):
                    with open(json_data_path, "r") as f_in:
                        json_data = json.load(f_in)
                        json_data["func_eval"] += new_function_evaluation_results
                    with open(json_data_path, "w") as f_out:
                        json.dump(json_data, f_out, indent=2)
            elif self.file_synchronization_method == 'rsync':
                while True:
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        json_data = json.load(f_in)
                        json_data["func_eval"] += new_function_evaluation_results
                    with open(temp_path, "w") as f_out:
                        json.dump(json_data, f_out, indent=2)
                    os.system("rsync -u " + temp_path + " " + json_data_path)
                    os.system("rm " + temp_path)
                    with open(json_data_path, "r") as f_in:
                        json_data = json.load(f_in)
                        existing_uids = [item["uid"] for item in json_data["func_eval"]]
                        new_uids = [item["uid"] for item in new_function_evaluation_results]
                        retry = False
                        for uid in new_uids:
                            if uid not in existing_uids:
                                retry = True
                                break
                        if retry == False:
                            break
            else:
                with open(json_data_path, "r") as f_in:
                    json_data = json.load(f_in)
                    json_data["func_eval"] += new_function_evaluation_results
                with open(json_data_path, "w") as f_out:
                    json.dump(json_data, f_out, indent=2)

        return

    def check_surrogate_model_exact_match(self,
            surrogate_model : dict,
            task_parameters_given: np.array,
            input_space_given: list,
            parameter_space_given: list,
            output_space_given: list):

        model_task_parameters = surrogate_model["task_parameters"]
        if len(model_task_parameters) != len(task_parameters_given):
            return False
        num_tasks = len(task_parameters_given)
        for i in range(num_tasks):
            if len(model_task_parameters[i]) != len(task_parameters_given[i]):
                return False
            for j in range(len(task_parameters_given[i])):
                if model_task_parameters[i][j] != task_parameters_given[i][j]:
                    return False

        for space_model in surrogate_model["input_space"]:
            space_given = next((item for item in input_space_given if item["name"] == space_model["name"]), None)
            if space_given == None:
                return False
            if space_model["type"] != space_given["type"]:
                return False
            if space_model["transformer"] != space_given["transformer"]:
                return False
            if space_model["lower_bound"] != space_given["lower_bound"]:
                return False
            if space_model["upper_bound"] != space_given["upper_bound"]:
                return False

        for space_model in surrogate_model["parameter_space"]:
            space_given = next((item for item in parameter_space_given if item["name"] == space_model["name"]), None)
            if space_given == None:
                return False
            if space_model["type"] != space_given["type"]:
                return False
            if space_model["transformer"] != space_given["transformer"]:
                return False
            if space_model["lower_bound"] != space_given["lower_bound"]:
                return False
            if space_model["upper_bound"] != space_given["upper_bound"]:
                return False

        for space_model in surrogate_model["output_space"]:
            space_given = next((item for item in output_space_given if item["name"] == space_model["name"]), None)
            if space_given == None:
                return False
            if space_model["type"] != space_given["type"]:
                return False
            if space_model["transformer"] != space_given["transformer"]:
                return False
            if space_model["lower_bound"] != space_given["lower_bound"]:
                return False
            if space_model["upper_bound"] != space_given["upper_bound"]:
                return False

        return True

    def check_surrogate_model_usable(self,
            surrogate_model : dict,
            task_parameters_given: np.array,
            input_space_given: list,
            parameter_space_given: list,
            output_space_given: list):

        #print ("check_surrogate_model_usable")
        #print ("surrogate_model: ", surrogate_model)
        #print ("task_parameters_given: ", task_parameters_given)
        #print ("input_space_given: ", input_space_given)
        #print ("parameter_space_given: ", parameter_space_given)
        #print ("output_space_given: ", output_space_given)

        model_task_parameters = surrogate_model["task_parameters"]
        task_parameter_names_given = [item["name"] for item in input_space_given]
        task_parameter_names_model = [item["name"] for item in surrogate_model["input_space"]]
        for input_task in task_parameters_given:
            input_task_ordered = [
                    input_task[task_parameter_names_given.index(name)]
                    for name in task_parameter_names_model]
            if not input_task_ordered in model_task_parameters:
                print ("[HistoryDB] Task information is not found")
                return False

        for space_model in surrogate_model["input_space"]:
            space_given = next((item for item in input_space_given if item["name"] == space_model["name"]), None)
            if space_given == None:
                return False
            if space_model["type"] != space_given["type"]:
                return False
            if space_model["transformer"] != space_given["transformer"]:
                return False
            #if space_model["lower_bound"] != space_given["lower_bound"]:
            #    return False
            #if space_model["upper_bound"] != space_given["upper_bound"]:
            #    return False

        for space_model in surrogate_model["parameter_space"]:
            space_given = next((item for item in parameter_space_given if item["name"] == space_model["name"]), None)
            if space_given == None:
                return False
            if space_model["type"] != space_given["type"]:
                return False
            if space_model["transformer"] != space_given["transformer"]:
                return False
            #if space_model["lower_bound"] != space_given["lower_bound"]:
            #    return False
            #if space_model["upper_bound"] != space_given["upper_bound"]:
            #    return False

        for space_model in surrogate_model["output_space"]:
            space_given = next((item for item in output_space_given if item["name"] == space_model["name"]), None)
            if space_given == None:
                return False
            if space_model["type"] != space_given["type"]:
                return False
            if space_model["transformer"] != space_given["transformer"]:
                return False
            #if space_model["lower_bound"] != space_given["lower_bound"]:
            #    return False
            #if space_model["upper_bound"] != space_given["upper_bound"]:
            #    return False

        return True

    def read_surrogate_models(self, tuningproblem=None, Igiven=None, modeler="Model_LCM"):
        ret = []
        print ("problem ", tuningproblem)
        print ("problem input_space ", self.problem_space_to_dict(tuningproblem.input_space))

        if tuningproblem == "None" or Igiven == "None":
            return ret

        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"
            if os.path.exists(json_data_path):
                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "r") as f_in:
                            history_data = json.load(f_in)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        history_data = json.load(f_in)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "r") as f_in:
                        history_data = json.load(f_in)

                num_models = len(history_data["surrogate_model"])

                max_evals = 0
                max_evals_index = -1 # TODO: if no model is found?
                for i in range(num_models):
                    surrogate_model = history_data["surrogate_model"][i]
                    if (self.check_surrogate_model_exact_match(
                        surrogate_model,
                        Igiven,
                        self.problem_space_to_dict(tuningproblem.input_space),
                        self.problem_space_to_dict(tuningproblem.parameter_space),
                        self.problem_space_to_dict(tuningproblem.output_space)) and
                        surrogate_model["modeler"] == modeler):
                        ret.append(surrogate_model)

        return ret

    def load_MLE_surrogate_model_hyperparameters(self, tuningproblem : TuningProblem,
            input_given : np.ndarray, objective : int, modeler : str):
        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"
            if os.path.exists(json_data_path):
                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "r") as f_in:
                            history_data = json.load(f_in)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        history_data = json.load(f_in)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "r") as f_in:
                        history_data = json.load(f_in)

                max_mle = -9999
                max_mle_index = -1
                for i in range(len(history_data["surrogate_model"])):
                    surrogate_model = history_data["surrogate_model"][i]
                    if (self.check_surrogate_model_exact_match(
                        surrogate_model,
                        input_given,
                        self.problem_space_to_dict(tuningproblem.input_space),
                        self.problem_space_to_dict(tuningproblem.parameter_space),
                        self.problem_space_to_dict(tuningproblem.output_space)) and
                        surrogate_model["modeler"] == modeler and
                        surrogate_model["objective_id"] == objective):
                        log_likelihood = surrogate_model["log_likelihood"]
                        if log_likelihood > max_mle:
                            max_mle = log_likelihood
                            max_mle_index = i
                if (max_mle_index == -1):
                    print ("Unable to find a model")
                    return None

                hyperparameters =\
                        history_data["surrogate_model"][max_mle_index]["hyperparameters"]

                parameter_names = []
                for parameter_info in history_db["surrogate_model"][max_mle_index]["parameter_space"]:
                    parameter_names.append(parameter_info["name"])

        return (hyperparameters, parameter_names)

    def load_AIC_surrogate_model_hyperparameters(self, tuningproblem : TuningProblem,
            input_given : np.ndarray, objective : int, modeler : str):
        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"
            if os.path.exists(json_data_path):
                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "r") as f_in:
                            history_data = json.load(f_in)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        history_data = json.load(f_in)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "r") as f_in:
                        history_data = json.load(f_in)

                min_aic = 99999
                min_aic_index = -1
                for i in range(len(history_data["surrogate_model"])):
                    surrogate_model = history_data["surrogate_model"][i]
                    if (self.check_surrogate_model_exact_match(
                        surrogate_model,
                        input_given,
                        self.problem_space_to_dict(tuningproblem.input_space),
                        self.problem_space_to_dict(tuningproblem.parameter_space),
                        self.problem_space_to_dict(tuningproblem.output_space)) and
                        surrogate_model["modeler"] == modeler and
                        surrogate_model["objective_id"] == objective):
                        log_likelihood = surrogate_model["log_likelihood"]
                        num_parameters = len(surrogate_model["hyperparameters"])
                        AIC = -1.0 * 2.0 * log_likelihood + 2.0 * num_parameters
                        if AIC < min_aic:
                            min_aic = AIC
                            min_aic_index = i
                if (min_aic_index == -1):
                    print ("Unable to find a model")
                    return None

                hyperparameters =\
                        history_data["surrogate_model"][min_aic_index]["hyperparameters"]

                parameter_names = []
                for parameter_info in history_db["surrogate_model"][min_aic_index]["parameter_space"]:
                    parameter_names.append(parameter_info["name"])

        return (hyperparameters, parameter_names)

    def load_BIC_surrogate_model_hyperparameters(self, tuningproblem : TuningProblem,
            input_given : np.ndarray, objective : int, modeler : str):
        import math

        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"
            if os.path.exists(json_data_path):
                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "r") as f_in:
                            history_data = json.load(f_in)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        history_data = json.load(f_in)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "r") as f_in:
                        history_data = json.load(f_in)

                min_bic = 99999
                min_bic_index = -1
                for i in range(len(history_data["surrogate_model"])):
                    surrogate_model = history_data["surrogate_model"][i]
                    if (self.check_surrogate_model_exact_match(
                        surrogate_model,
                        input_given,
                        self.problem_space_to_dict(tuningproblem.input_space),
                        self.problem_space_to_dict(tuningproblem.parameter_space),
                        self.problem_space_to_dict(tuningproblem.output_space)) and
                        surrogate_model["modeler"] == modeler and
                        surrogate_model["objective_id"] == objective):
                        log_likelihood = surrogate_model["log_likelihood"]
                        num_parameters = len(surrogate_model["hyperparameters"])
                        num_samples = len(surrogate_model["function_evaluations"])
                        BIC = -1.0 * 2.0 * log_likelihood + num_parameters * math.log(num_samples)
                        if BIC < min_bic:
                            min_bic = BIC
                            min_bic_index = i
                if (min_bic_index == -1):
                    print ("Unable to find a model")
                    return None

                hyperparameters =\
                        history_data["surrogate_model"][min_bic_index]["hyperparameters"]

                parameter_names = []
                for parameter_info in history_db["surrogate_model"][min_bic_index]["parameter_space"]:
                    parameter_names.append(parameter_info["name"])

        return (hyperparameters, parameter_names)

    def load_max_evals_surrogate_model_hyperparameters(self, tuningproblem : TuningProblem,
            input_given : np.ndarray, objective : int, modeler : str):
        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"
            if os.path.exists(json_data_path):
                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "r") as f_in:
                            history_data = json.load(f_in)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        history_data = json.load(f_in)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "r") as f_in:
                        history_data = json.load(f_in)

                max_evals = 0
                max_evals_index = -1 # TODO: if no model is found?
                for i in range(len(history_data["surrogate_model"])):
                    surrogate_model = history_data["surrogate_model"][i]
                    if (self.check_surrogate_model_exact_match(
                        surrogate_model,
                        input_given,
                        self.problem_space_to_dict(tuningproblem.input_space),
                        self.problem_space_to_dict(tuningproblem.parameter_space),
                        self.problem_space_to_dict(tuningproblem.output_space)) and
                        surrogate_model["modeler"] == modeler and
                        surrogate_model["objective_id"] == objective):
                        num_evals = len(history_data["surrogate_model"][i]["function_evaluations"])
                        if num_evals > max_evals:
                            max_evals = num_evals
                            max_evals_index = i
                if (max_evals_index == -1):
                    print ("Unable to find a model")
                    return None

                hyperparameters =\
                        history_data["surrogate_model"][max_evals_index]["hyperparameters"]

                parameter_names = []
                for parameter_info in history_data["surrogate_model"][max_evals_index]["parameter_space"]:
                    parameter_names.append(parameter_info["name"])

        return (hyperparameters, parameter_names)

    def load_surrogate_model_hyperparameters_by_uid(self, model_uid):
        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"
            if os.path.exists(json_data_path):
                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "r") as f_in:
                            history_data = json.load(f_in)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        history_data = json.load(f_in)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "r") as f_in:
                        history_data = json.load(f_in)

                surrogate_model = history_data["surrogate_model"]
                num_models = len(surrogate_model)
                for i in range(num_models):
                    if surrogate_model[i]["uid"] == model_uid:
                        return surrogate_model[i]["hyperparameters"]

        return []

    def load_surrogate_model_meta_data(self,
            task_parameters_given: np.ndarray,
            tuning_configuration: dict,
            input_space_given: dict,
            parameter_space_given: dict,
            output_space_given: dict,
            objective: int,
            modeler: str):
            #tuningproblem : TuningProblem,
            #input_given : np.ndarray, objective : int, modeler : str):
        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"
            if os.path.exists(json_data_path):
                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "r") as f_in:
                            history_data = json.load(f_in)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        history_data = json.load(f_in)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "r") as f_in:
                        history_data = json.load(f_in)

                max_evals = 0
                max_evals_index = -1 # TODO: if no model is found?
                for i in range(len(history_data["surrogate_model"])):
                    surrogate_model = history_data["surrogate_model"][i]
                    if (self.check_surrogate_model_usable(surrogate_model,
                        task_parameters_given,
                        input_space_given,
                        parameter_space_given,
                        output_space_given) and
                        surrogate_model["modeler"] == modeler and
                        surrogate_model["objective_id"] == objective):

                        tuning_configuration_match = True
                        for func_eval_uid in surrogate_model["function_evaluations"]:
                            func_eval = search_item_by_uid(history_data["func_eval"], func_eval_uid)
                            #print ("tuning_configuration (machine): ", tuning_configuration["machine_configuration"])
                            #print ("func_eval (machine):            ", func_eval["machine_configuration"])
                            if tuning_configuration != None:
                                if str(tuning_configuration["machine_configuration"]) != str(func_eval["machine_configuration"]):
                                    #print ("not same")
                                    tuning_configuration_match = False
                                    break
                                #else:
                                #    print ("same")

                            #if tuning_configuration != None and tuning_configuration["software_configuration"] != func_eval["software_configuration"]:
                            #    tuning_configuration_match = False
                            #    break
                        if tuning_configuration_match:
                            num_evals = len(history_data["surrogate_model"][i]["function_evaluations"])
                            if num_evals > max_evals:
                                max_evals = num_evals
                                max_evals_index = i
                if (max_evals_index == -1):
                    print ("Unable to find a surrogate model")
                    return None

        return history_data["surrogate_model"][max_evals_index]

    def load_surrogate_model_configurations(self,
            task_parameters_given: np.ndarray,
            input_space_given: dict,
            parameter_space_given: dict,
            output_space_given: dict,
            loadable_machine_configurations: dict,
            loadable_software_configurations: dict,
            objective: int,
            modeler: str):

        model_configurations = []

        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"
            if os.path.exists(json_data_path):
                if self.file_synchronization_method == 'filelock':
                    with FileLock(json_data_path+".lock"):
                        with open(json_data_path, "r") as f_in:
                            history_data = json.load(f_in)
                elif self.file_synchronization_method == 'rsync':
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        history_data = json.load(f_in)
                    os.system("rm " + temp_path)
                else:
                    with open(json_data_path, "r") as f_in:
                        history_data = json.load(f_in)

                for surrogate_model in history_data["surrogate_model"]:

                    if (self.check_surrogate_model_usable(surrogate_model,
                        task_parameters_given,
                        input_space_given,
                        parameter_space_given,
                        output_space_given) and
                        surrogate_model["modeler"] == modeler and
                        surrogate_model["objective_id"] == objective):
                        #print (surrogate_model)

                        for func_eval_uid in surrogate_model["function_evaluations"]:
                            func_eval = search_item_by_uid(history_data["func_eval"], func_eval_uid)

                            configuration = {
                                    "task_parameters": surrogate_model["task_parameters"],
                                    "machine_configuration": func_eval["machine_configuration"],
                                    "software_configuration": func_eval["software_configuration"]
                                    }

                            if configuration not in model_configurations:
                                model_configurations.append(configuration)

        return (model_configurations)

    def problem_space_to_dict(self, space : Space):

        dict_arr = []

        space_len = len(space)

        transformers = space.get_transformer()

        for i in range(space_len):
            dict_ = {}

            dict_["name"] = space[i].name

            dict_["transformer"] = transformers[i]

            space_type_name = type(space[i]).__name__

            if space_type_name == "Real":
                dict_["type"] = "real"

                lower_bound, upper_bound = space.bounds[i]
                dict_["lower_bound"] = lower_bound
                dict_["upper_bound"] = upper_bound

                dict_["optimize"] = space[i].optimize

            elif space_type_name == "Integer":
                dict_["type"] = "int"

                lower_bound, upper_bound = space.bounds[i]
                dict_["lower_bound"] = lower_bound
                dict_["upper_bound"] = upper_bound

                dict_["optimize"] = space[i].optimize

            elif space_type_name == "Categoricalnorm":
                dict_["type"] = "categorical"
                dict_["categories"] = space[i].bounds

            else:
                print ("space type unknown")

            dict_arr.append(dict_)

        return dict_arr

    def store_model_LCM(self,\
            objective : int,
            problem : Problem,\
            input_given : np.ndarray,\
            bestxopt : np.ndarray,\
            neg_log_marginal_likelihood : float,\
            gradients : np.ndarray,\
            iteration : int):

        if (self.tuning_problem_name is not None):
            json_data_path = self.history_db_path+"/"+self.tuning_problem_name+".json"

            new_surrogate_models = []

            now = time.localtime()

            #from scipy.stats.mstats import gmean
            #from scipy.stats.mstats import hmean
            model_stats = {}
            model_stats["log_likelihood"] = -1.0 * neg_log_marginal_likelihood
            model_stats["neg_log_likelihood"] = neg_log_marginal_likelihood
            model_stats["gradients"] = gradients.tolist()
            #model_stats["gradients_sum_abs"] = np.sum(np.absolute(gradients))
            #model_stats["gradients_average_abs"] = np.average(np.absolute(gradients))
            #model_stats["gradients_hmean_abs"] = hmean(np.absolute(gradients))
            #model_stats["gradients_gmean_abs"] = gmean(np.absolute(gradients))
            model_stats["iterations"] = iteration

            gradients_list = gradients.tolist()

            #problem_space = {}
            #problem_space["input_space"] = self.problem_space_to_dict(problem.IS)
            #problem_space["parameter_space"] = self.problem_space_to_dict(problem.PS)
            #problem_space["output_space"] = self.problem_space_to_dict(problem.OS)

            task_parameter_orig = problem.IS.inverse_transform(np.array(input_given, ndmin=2))
            task_parameter_orig_list = np.array(task_parameter_orig).tolist()
            #task_parameter_dict_list = []
            #for i in range(len(input_given)):
            #    task_parameter_dict = {problem.IS[k].name:task_parameter_orig_list[i][k] for k in range(len(problem.IS))}
            #    task_parameter_dict_list.append(task_parameter_dict)

            objective_dict = self.problem_space_to_dict(problem.OS)[objective]
            objective_dict["objective_id"] = objective

            new_surrogate_models.append({
                    "hyperparameters":bestxopt.tolist(),
                    "model_stats":model_stats,
                    "task_parameters":task_parameter_orig_list,
                    "function_evaluations":self.uids,
                    "input_space":self.problem_space_to_dict(problem.IS),
                    "parameter_space":self.problem_space_to_dict(problem.PS),
                    "output_space":self.problem_space_to_dict(problem.OS),
                    "modeler":"Model_LCM",
                    "objective":objective_dict,
                    "time":{
                        "tm_year":now.tm_year,
                        "tm_mon":now.tm_mon,
                        "tm_mday":now.tm_mday,
                        "tm_hour":now.tm_hour,
                        "tm_min":now.tm_min,
                        "tm_sec":now.tm_sec,
                        "tm_wday":now.tm_wday,
                        "tm_yday":now.tm_yday,
                        "tm_isdst":now.tm_isdst
                        },
                    "uid":str(uuid.uuid1())
                    # objective id is to dinstinguish between different models for multi-objective optimization;
                    # we might need a nicer way to manage different models
                })

            if self.file_synchronization_method == 'filelock':
                with FileLock(json_data_path+".lock"):
                    with open(json_data_path, "r") as f_in:
                        json_data = json.load(f_in)
                        json_data["surrogate_model"] += new_surrogate_models
                    with open(json_data_path, "w") as f_out:
                        json.dump(json_data, f_out, indent=2)
            elif self.file_synchronization_method == 'rsync':
                while True:
                    temp_path = json_data_path + "." + self.process_uid + ".temp"
                    os.system("rsync -a " + json_data_path + " " + temp_path)
                    with open(temp_path, "r") as f_in:
                        json_data = json.load(f_in)
                        json_data["surrogate_model"] += new_surrogate_models
                    with open(temp_path, "w") as f_out:
                        json.dump(json_data, f_out, indent=2)
                    os.system("rsync -u " + temp_path + " " + json_data_path)
                    os.system("rm " + temp_path)
                    with open(json_data_path, "r") as f_in:
                        json_data = json.load(f_in)
                        existing_uids = [item["uid"] for item in json_data["surrogate_model"]]
                        new_uids = [item["uid"] for item in new_surrogate_models]
                        retry = False
                        for uid in new_uids:
                            if uid not in existing_uids:
                                retry = True
                                break
                        if retry == False:
                            break
            else:
                with open(json_data_path, "r") as f_in:
                    json_data = json.load(f_in)
                    json_data["surrogate_model"] += new_surrogate_models
                with open(json_data_path, "w") as f_out:
                    json.dump(json_data, f_out, indent=2)

        return



# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#
import numpy as np
from autotune.problem import TuningProblem
from computer import Computer
from typing import Collection

import skopt.space
from skopt.space import *
from data import *
import math
import opentuner
import argparse
import functools
import time

####################################################################################################

class OpenTunerInterface(opentuner.MeasurementInterface):

    def __init__(self, args):

        super(OpenTunerInterface, self).__init__(args)
        self.args = args
        self.X = []
        self.Y = []
        self.timefun = 0

    def manipulator(self):

        """
        Define the search space by creating a ConfigurationManipulator
        """

        manipulator = opentuner.ConfigurationManipulator()
        for n,p in zip(self.args.tp.parameter_space.dimension_names,self.args.tp.parameter_space.dimensions):
            if (isinstance(p, Real)):
                manipulator.add_parameter(opentuner.FloatParameter(n, p.bounds[0], p.bounds[1]))
            elif (isinstance(p, Integer)):
                manipulator.add_parameter(opentuner.IntegerParameter(n, p.bounds[0], p.bounds[1]))
            elif (isinstance(p, Categorical)):
                manipulator.add_parameter(opentuner.search.manipulator.EnumParameter(n, list(p.bounds)))
            else:
                raise Exception("Unknown parameter type")

        self.cpt = 0

        return manipulator

    def run(self, desired_result, input, limit):

        """
        Run a given configuration then return performance
        """
        t1 = time.time_ns()
        # Extract parameters

        cfg = desired_result.configuration.data
        t = self.args.t
        x = [cfg[p] for p in self.args.tp.parameter_space.dimension_names]
        kwargs = {d.name: x[i] for (i, d) in enumerate(self.args.tp.parameter_space)}
        kwargs2 = {d.name: t[i] for (i, d) in enumerate(self.args.tp.input_space)}
        kwargs2.update(kwargs)
        check_constraints = functools.partial(self.args.computer.evaluate_constraints, self.args.tp, inputs_only = False, kwargs = kwargs)
        cond = check_constraints(kwargs2)
        if (cond):
            y = float(self.args.tp.objective(kwargs2)[0])
            self.X.append(x)
            self.Y.append(y)
            state = 'OK'
            self.cpt += 1
            print(t, x, y, state, self.args.test_limit, self.cpt)
            # sys.stdout.flush()
        else:
            y = float("Inf")
            state = 'ERROR'
#            self.args.test_limit = min(self.args.test_limit + 1, 1000) #XXX Hack
            self.args.test_limit = self.args.test_limit + 1 #XXX Hack
        t2 = time.time_ns()
        self.timefun=self.timefun+(t2-t1)/1e9

        return opentuner.Result(time=y, state=state)

    def save_final_config(self, configuration):

        """called at the end of tuning"""

        self.Xopt = np.array([configuration.data[p] for p in self.args.tp.parameter_space.dimension_names])
        self.Yopt = min(self.Y)

####################################################################################################

def OpenTuner(T, NS, tp : TuningProblem, computer : Computer, run_id="OpenTuner", niter=1, technique=None):

    # Initialize

    args = argparse.Namespace()

    args.bail_threshold            = 500
    args.database                  = None
    args.display_frequency         = 10
    args.generate_bandit_technique = False
    args.label                     = run_id
    args.list_techniques           = False
    args.machine_class             = None
    args.no_dups                   = True
    args.parallel_compile          = False
    args.parallelism               = 1
    args.pipelining                = 0
    args.print_params              = False
    args.print_search_space_size   = False
    args.quiet                     = True
    args.results_log               = None
    args.results_log_details       = None
    args.seed_configuration        = []
    args.stop_after                = None
    args.technique                 = technique
    args.test_limit                = NS

    args.niter   = niter
    args.tp  = tp
    args.computer  = computer

    X = []
    Y = []
    data = Data(tp)
    # Xopt = []
    # Yopt = []

    # Tune
    stats = {
        "time_total": 0,
        "time_fun": 0
    }

    timefun=0
    t1 = time.time_ns()
    print("Start OpenTuner")
    for i in range(len(T)):

        args.t = T[i]
        args.test_limit = NS #XXX Hack
        a = OpenTunerInterface(args)
        b = opentuner.tuningrunmain.TuningRunMain(a, args)
        b.main()

        X.append(a.X)
        tmp = np.array(a.Y).reshape((len(a.Y), 1))
        Y.append(tmp)
        timefun=timefun+a.timefun
        # Xopt.append(a.Xopt)
        # Yopt.append(a.Yopt)

    print("End OpenTuner")

    t2 = time.time_ns()
    stats['time_total'] = (t2-t1)/1e9
    stats['time_fun'] = timefun

    data.I=T
    data.P=X
    data.O=Y
    # Finalize

    return (data, stats)




# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.
#
import numpy as np
from problem import Problem
from typing import Collection

import skopt.space
from skopt.space import *
import math

class Categoricalnorm(Categorical):
    def transform(self, X):
        lens=len(self.categories)
        Xt = super(Categoricalnorm, self).transform(X)
        tmp1=[]
        for xt in Xt:
            if(len(xt)==1):
                tmp = min(1. - 1e-12, xt[0]+1e-12)
            else:
                ii = next(i for i,v in enumerate(xt) if v > 0)
                tmp = ii/lens+0.01
            tmp1.append(tmp)
        return tmp1

    def inverse_transform(self, Xt):
        lens=len(self.categories)
        tmp1=[]
        # print(Xt,'wideeeefd')
        for xt in Xt:
            # print(xt,'widfd')
            tmp=[0 for ii in range(lens)]
            tmp[math.floor(min(xt,1-1e-12)*lens)]=1
            tmp1.append(tmp)
        # print(tmp1,'before inverse_transform',Xt,xt*lens)

        return super(Categoricalnorm, self).inverse_transform(tmp1)

    @property
    def transformed_size(self):
        return 1

class Data(object):
    # To GPTune I is 2D numpy array. To user I is a list of lists
    # To GPTune P is a list/collection of 2D numpy array with column dimension corresponding to PS dimension. To user P is a list of (list of lists)
    # To GPTune and user O is a list/collection of 2D numpy array with column dimension 1 for single-objective function.
    # To GPTune and user D is a list/collection of dictionaries
    def __init__(self, problem : Problem, I = None, P = None, O = None, D = None):
    # def __init__(self, problem : Problem, I : np.ndarray = None, P : Collection[np.ndarray] = None, O : Collection[np.ndarray] = None):

        self.problem = problem

        # if (not self.check_inputs(I)):
        #     raise Exception("")

        self.I = I

        # if (not self.check_parameters(P)):
        #     raise Exception("")

        self.P = P

        # if (not self.check_outputs(O)):
        #     raise Exception("")

        self.O = O


        self.D = D

    @property
    def NI(self):

        if (self.I is None):
            return 0
        else:
            return len(self.I)

    def check_inputs(self, I: np.ndarray) -> bool:

        cond = True
        if (I is not None):
            if (not (I.ndim == 2 and I.shape[1] == self.problem.DI)):
                cond = False

        return cond

    def check_parameters(self, P: Collection[np.ndarray]) -> bool:

        cond = True
        if (P is not None):
            for x in P:
                if (x is not None and len(x) > 0):
                    if not (x.ndim == 2 and x.shape[1] == problem.DP):
                        cond = False
                        break

        return cond

    def check_outputs(self, O: Collection[np.ndarray]) -> bool:

        cond = True
        if (O is not None):
            for o in O:
                if (o is not None and len(o) > 0):
                    if not (o.ndim == 2 and o.shape[1] == problem.DO):
                        cond = False
                        break

        return cond

    # TODO
    def points2kwargs(self):

        # transform the self.I and self.P into a list of dictionaries

        pass

    # TODO
    def merge(self, newdata):

        # merge the newdata with self, making sure that the Ts coincide

        if (not np.array_equal(self.I, newdata.I)):
            raise Exception("The tasks in the newdata should be the same as the current tasks")

        if (not np.array_equal(self.D, newdata.D)):
            raise Exception("The tasks dictionaries in the newdata should be the same as the current tasks")

        self.P = [np.concatenate((self.P[i], newdata.P[i])) for i in range(len(self.P))]
        self.O = [np.concatenate((self.O[i], newdata.O[i])) for i in range(len(self.O))]

#    def insert(I = None: np.ndarray, P = None : Collection[np.ndarray], O = None : Collection[np.ndarray]):
#
#        if (I is not None):
#            if (I.ndim == 1):
#                assert(I.shape[0] == self.problem.DI)
#            elif (I.ndim == 2):
#                assert(I.shape[1] == self.problem.DI)
#            else:
#                raise Exception("")
#            self.I.append(I)

class HistoricData(Data):

    def __init__(self, data, computer, options):

        self.data     = data
        self.computer = computer
        self.options  = options

#   @staticmethod
#   def load():
#
#    pass
#
#   def save(self):
#
#    pass




#! /usr/bin/env python


# GPTune Copyright (c) 2019, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory (subject to receipt of any
# required approvals from the U.S.Dept. of Energy) and the University of
# California, Berkeley.  All rights reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Intellectual Property Office at IPO@lbl.gov.
#
# NOTICE. This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights.
# As such, the U.S. Government has been granted for itself and others acting
# on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in
# the Software to reproduce, distribute copies to the public, prepare
# derivative works, and perform publicly and display publicly, and to permit
# other to do so.

import os, ctypes
import numpy as np
import GPy
import mpi4py
import itertools
import scipy
import sys
from sys import platform
import time

ROOTDIR = os.path.abspath(__file__ + "/../../build")

if platform == "linux" or platform == "linux2":
    cliblcm = ctypes.cdll.LoadLibrary(ROOTDIR + '/lib_gptuneclcm.so')
elif platform == "darwin":
    cliblcm = ctypes.cdll.LoadLibrary(ROOTDIR + '/lib_gptuneclcm.dylib')
elif platform == "win32":
    raise Exception(f"Windows is not yet supported")





####################################################################################################

class LCM(GPy.kern.Kern):

    """
    LCM kernel:

    .. math::

    """

    def __init__(self, input_dim, num_outputs, Q, name='LCM'):  # self and input_dim are required for GPy

        super(LCM, self).__init__(input_dim + 1, active_dims=None, name=name)

        self.num_outputs = num_outputs
        self.Q = Q

        self.theta =       np.power(10,np.random.randn(Q * input_dim))
        self.var   =       np.power(10,np.random.randn(Q))
        self.kappa =       np.power(10,np.random.randn(Q * num_outputs))
        self.sigma =       np.power(10,np.random.randn(num_outputs))
        self.WS    =       np.power(10,np.random.randn(Q * num_outputs))
        # print('why????',self.theta,self.var,self.kappa,self.sigma,self.WS)

    #     self.theta =  0.54132485 * np.ones(Q * input_dim)
    #     self.var   =  0.54132485 * np.ones(Q)
    #     self.kappa = -0.43275213 * np.ones(Q * num_outputs)
    #     self.sigma =  0.54132485 * np.ones(num_outputs)
    # #        np.random.seed(0)
    #     self.WS    =   .5 * np.random.randn(Q * num_outputs)

        self.BS    = np.empty(Q * self.num_outputs ** 2)

    def get_param_array(self):

        x = np.concatenate([self.theta, self.var, self.kappa, self.sigma, self.WS])

        return x


    def get_correlation_metric(self):
        # self.kappa =  b_{1,1}, ..., b_{delta,1}, ..., b_{1,Q}, ..., b_{\delta,Q}
        # self.sigma = d_1, ..., d_delta
        # self.WS = a_{1,1}, ..., a_{delta,1}, ..., a_{1,Q}, ..., a_{delta,Q}
        kappa = self.kappa
        sigma = self.sigma
        WS = self.WS
        delta = len(sigma)
        Q = int(len(WS)/delta)
        # print('NI = ', delta)
        # print('Q = ', Q)
        B = np.zeros((delta, delta, Q))
        for i in range(Q):
            Wq = WS[i*delta : (i+1)*delta]
            Kappa_q = kappa[i*delta : (i+1)*delta]
            B[:, :, i] = np.outer(Wq, Wq) + np.diag(Kappa_q)
            # print("In model.py, i = ", i)
            # print(B[:, :, i])

        # return C_{i, i'}
        C = np.zeros((delta, delta))
        for i in range(delta):
            for ip in range(i, delta):
                C[i, ip] = np.linalg.norm(B[i, ip, :]) / np.sqrt(np.linalg.norm(B[i, i, :]) * np.linalg.norm(B[ip, ip, :]))
        return C


    def set_param_array(self, x):

        cpt = 0
        for i in range(len(self.theta)):
            self.theta[i] = x[cpt]
            cpt += 1
        for i in range(len(self.var)):
            self.var[i] = x[cpt]
            cpt += 1
        for i in range(len(self.kappa)):
            self.kappa[i] = x[cpt]
            cpt += 1
        for i in range(len(self.sigma)):
            self.sigma[i] = x[cpt]
            cpt += 1
        for i in range(len(self.WS)):
            self.WS[i] = x[cpt]
            cpt += 1

        self.parameters_changed()

        # print(self.theta)
        # print(self.var)
        # print(self.kappa)
        # print(self.sigma)
        # print(self.WS)
        # print(self.BS)

    def parameters_changed(self):

        for q in range(self.Q):

            ws = self.WS[q * self.num_outputs : (q + 1) * self.num_outputs].reshape(1, self.num_outputs)
            a = np.dot(ws.T, ws) + np.diag(self.kappa[q * self.num_outputs : (q + 1) * self.num_outputs])
            self.BS[q * self.num_outputs ** 2 : (q + 1) * self.num_outputs ** 2] = a.flatten()

    def K(self, X1, X2=None):   # Required for GPy, X1 and X2 are ndarray stored in row major

        if X2 is None: X2 = X1
        # print("cao",X1)
        K = np.empty((X1.shape[0], X2.shape[0]))

        try:
            cliblcm.K(ctypes.c_int(self.input_dim - 1),\
                    ctypes.c_int(self.num_outputs),\
                    ctypes.c_int(self.Q),\
                    ctypes.c_int(X1.shape[0]),\
                    ctypes.c_int(X2.shape[0]),\
                    self.theta.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),\
                    self.var.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),\
                    self.BS.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),\
                    X1.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),\
                    X2.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),\
                    K.ctypes.data_as(ctypes.POINTER(ctypes.c_double)))
        except Exception as inst:
            print(inst)

        # print("cadfdfo",X1)
        return K

    def Kdiag(self, X):   # Required for GPy

        return np.diag(self.K(X, X2=X))

    def update_gradients_full(self, dL_dK, X1, X2=None):

        pass

    def update_gradients_diag(self, dL_dKdiag, X):

        pass

    def gradients_X(self, dL_dK, X1, X2):

        raise("Not implemented")

    def gradients_X_diag(self,dL_dKdiag,X):

        raise("Not implemented")

    def train_kernel(self, X, Y, computer, kwargs):
        npernode = int(computer.cores/kwargs['model_threads'])
        maxtries = kwargs['model_max_jitter_try']
        jitter = kwargs['model_jitter']
        mpi_size=kwargs['model_processes']  # this makes sure every rank belongs to the blacs grid
        nprow = int(np.sqrt(mpi_size))
        npcol = mpi_size // nprow
        mpi_size = nprow * npcol

        t1 = time.time_ns()
        mpi_comm = computer.spawn(__file__, nproc=mpi_size, nthreads=kwargs['model_threads'], npernode=npernode, kwargs = kwargs)
        t2 = time.time_ns()
        if (kwargs['verbose']):
            print('LCM spawn time: ',(t2-t1)/1e9)

        X = np.concatenate([np.concatenate([X[i], np.ones((len(X[i]), 1)) * i], axis=1) for i in range(len(X))])
        Y = np.array(list(itertools.chain.from_iterable(Y)))

        _ = mpi_comm.bcast(("init", (self, X, Y, maxtries,jitter)), root=mpi4py.MPI.ROOT)

        _log_lim_val = np.log(np.finfo(np.float64).max)
        _exp_lim_val = np.finfo(np.float64).max
        _lim_val = 36.0
        epsilon = np.finfo(np.float64).resolution

        def transform_x(x):  # YL: Why is this needed?

            x2 = np.power(10,x.copy())

            # x2[list(range(len(self.theta)+len(self.var)+len(self.kappa)+len(self.sigma),len(x0)))] = np.log(x2[list(range(len(self.theta)+len(self.var)+len(self.kappa)+len(self.sigma),len(x)))])

            # for i in range(len(self.theta) + len(self.var) + len(self.kappa) + len(self.sigma)):
            #     x2[i] = np.where(x[i]>_lim_val, x[i], np.log1p(np.exp(np.clip(x[i], -_log_lim_val, _lim_val)))) #+ epsilon
            #     #x2[i] = np.where(x[i]>_lim_val, x[i], np.log(np.expm1(x[i]))) #+ epsilon

            return x2

        def inverse_transform_x(x):  # YL: Why is this needed?

            x0 = x.copy()
            ws = x0[list(range(len(self.theta)+len(self.var)+len(self.kappa)+len(self.sigma),len(x0)))]
            x2 = np.log10(x0)
            # x2[list(range(len(self.theta)+len(self.var)+len(self.kappa)+len(self.sigma),len(x0)))] = ws
            return x2



        def transform_gradient(x, grad):  # YL: Why is this needed?

            grad2 = grad.copy()
            # x2 = transform_x(x)
            # for i in range(len(self.theta) + len(self.var) + len(self.kappa) + len(self.sigma)):
            #     grad2[i] = grad[i]*np.where(x2[i]>_lim_val, 1., - np.expm1(-x2[i]))

            return grad2

        # Gradient-based optimization

        gradients = np.zeros(len(self.theta) + len(self.var) + len(self.kappa) + len(self.sigma) + len(self.WS))
        iteration = [0] #np.array([0])

        history_xs = [None]
        history_fs = [float('Inf')]

        def fun(x, *args):

            # print(np.power(10,x),'hp')
            t3 = time.time_ns()
            x2 = transform_x(x)
            # x2 = np.insert(x2,len(self.theta), np.ones(len(self.var)))  # fix self.var to 1
            _ = mpi_comm.bcast(("fun_jac", x2), root=mpi4py.MPI.ROOT)
    #            gradients[:] = 0.
            # print("~~~~")
            (neg_log_marginal_likelihood, g) = mpi_comm.recv(source = 0)
            # print("@@@@")
            # print(x2,neg_log_marginal_likelihood)
            #print ("g: ", g)
            #print ("iteration: " + str(iteration[0]))

            iteration[0] += 1

            gradients[:] = g[:]
            if (kwargs['verbose']):
                sys.stdout.flush()
            if (neg_log_marginal_likelihood < min(history_fs)):
                history_xs.append(x2)
                history_fs.append(neg_log_marginal_likelihood)
            t4 = time.time_ns()
            # print('fun_jac py: ',(t4-t3)/1e9)

            return (neg_log_marginal_likelihood)

        def grad(x, *args):
            # x = np.insert(x,len(self.theta), np.ones(len(self.var))) # fix self.var to 1
            grad = - gradients
            grad = transform_gradient(x, grad)

            # grad = np.delete(grad,list(range(len(self.theta),len(self.theta)+len(self.var)))) # fix self.var to 1
            return (grad)

        x0 = self.get_param_array()
        x0_log = inverse_transform_x(x0)

        # x0_log[0]=0
        x0_log[list(range(len(self.theta),len(self.theta)+len(self.var)))]=0
        # x0_log[2]=0
        # x0_log[3]=-10
        # x0_log[4]=-10
        # print(x0_log,'before')
        # sol = scipy.optimize.show_options(method='L-BFGS-B', disp=True, solver='minimize')
        t3 = time.time_ns()

        # bounds = [(-10, 10)] * len(x0_log)
        bounds = [(-10, 8)] * len(self.theta) + [(None, None)] * len(self.var) + [(-10, 8)] * len(self.kappa)+ [(-10, -5)] * len(self.sigma)+ [(-10, 6)] * len(self.WS)
        # print(bounds)

        # sol = scipy.optimize.minimize(fun, x0_log, args=(), method='L-BFGS-B', jac=grad)
        sol = scipy.optimize.minimize(fun, x0_log, args=(), method='L-BFGS-B', jac=grad, bounds=bounds, tol=None, callback=None, options={'disp': None, 'maxcor': 10, 'ftol': 1e-32, 'gtol': 1e-05, 'eps': 1e-08, 'maxfun': 1000, 'maxiter': 1000, 'iprint': -1, 'maxls': 100})

        # print(sol.x,'after')
        # print(transform_x(sol.x),'after exp')  # sol.x is not yet transformed

        t4 = time.time_ns()
        if (kwargs['verbose']):
            print('L-BFGS time: ',(t4-t3)/1e9)

        if (kwargs['verbose']):
            print('fun      : ', sol.fun)
            #print('hess_inv : ', sol.hess_inv)
            #print('jac      : ', jac)
            print('message  : ', sol.message)
            print('nfev     : ', sol.nfev)
            print('nit      : ', sol.nit)
            print('status   : ', sol.status)
            print('success  : ', sol.success)
            #print('x        : ', x)
    #            xopt = transform_x(sol.x)
    #            fopt = sol.fun
        xopt = history_xs[history_fs.index(min(history_fs))] # history_xs is already transformed
        fopt = min(history_fs)
        #print ("gradients: ", str(gradients))
        #print ("iteration: " + str(iteration[0]))

        if(xopt is None):
            raise Exception(f"L-BFGS failed: consider reducing options['model_latent'] !")


    #        # Particle Swarm Optimization
    #
    #        import pyswarms as ps
    #        min_bound = np.array([self.bounds[i][0] for i in range(len(self.bounds))], dtype='float64')
    #        max_bound = np.array([self.bounds[i][1] for i in range(len(self.bounds))], dtype='float64')
    #        bounds = (min_bound, max_bound)
    #        optimizer = ps.single.GlobalBestPSO(n_particles=100, dimensions=len(self.bounds), options={'c1': 0.5, 'c2': 0.3, 'w': 0.9}, bounds=bounds)
    #        fopt, xopt = optimizer.optimize(fun, iters=100)
    #        xopt = transform_x(xopt)
    #
    #        import pyswarm
    #        min_bound = np.array([-20 for i in range(len(self.bounds))], dtype='float64')
    #        max_bound = np.array([ 20 for i in range(len(self.bounds))], dtype='float64')
    #        xopt, fopt = pyswarm.pso(fun, min_bound, max_bound, ieqcons=[], f_ieqcons=None, args=(), kwargs={}, swarmsize=100, omega=0.5, phip=0.5, phig=0.5, maxiter=100, minstep=1e-8, minfunc=1e-8, debug=False)
    #        xopt = transform_x(xopt)

        self.set_param_array(xopt)
        _ = mpi_comm.bcast(("end", None), root=mpi4py.MPI.ROOT)

        mpi_comm.Disconnect()

        return (xopt, fopt, gradients, iteration[0])

if __name__ == "__main__":

    from ctypes import Structure, c_int, c_double, c_void_p, POINTER
    from mpi4py import MPI
    if mpi4py.MPI._sizeof(mpi4py.MPI.Comm) == ctypes.sizeof(ctypes.c_int):
        c_mpi_comm_t = c_int
    else:
        c_mpi_comm_t = c_void_p

    class fun_jac_struct(Structure):
        _fields_ = [("DI", c_int),\
                    ("NT", c_int),\
                    ("NL", c_int),\
                    ("nparam", c_int),\
                    ("m" , c_int),\
                    ("X", POINTER(c_double)),\
                    ("Y", POINTER(c_double)),\
                    ("dists", POINTER(c_double)),\
                    ("exps", POINTER(c_double)),\
                    ("alpha", POINTER(c_double)),\
                    ("K", POINTER(c_double)),\
                    ("gradients_TPS", POINTER(POINTER(c_double))),\
                    ("mb", c_int),\
                    ("lr", c_int),\
                    ("lc", c_int),\
                    ("maxtries", c_int),\
                    ("nprow", c_int),\
                    ("npcol", c_int),\
                    ("pid", c_int),\
                    ("prowid", c_int),\
                    ("pcolid", c_int),\
                    ("context", c_int),\
                    ("Kdesc", POINTER(c_int)),\
                    ("alphadesc", POINTER(c_int)),\
                    ("jitter", c_double),\
                    ("distY", POINTER(c_double)),\
                    ("buffer", POINTER(c_double)),\
                    ("mpi_comm", POINTER(c_mpi_comm_t))]

    mpi_comm = mpi4py.MPI.Comm.Get_parent()
    #    mpi_comm.Merge()

    #    color = self.mpi_rank // (self.mpi_size // num_subgroups)
    #    key   = self.mpi_rank %  (self.mpi_size // num_subgroups)
    #
    mpi_size = mpi_comm.Get_size()
    mpi_rank = mpi_comm.Get_rank()
    nprow = int(np.sqrt(mpi_size))
    npcol = mpi_size // nprow
    #    assert(nprow * npcol == mpi_size)
    mb = 32

    cond = True
    while (cond):

        res = mpi_comm.bcast(None, root=0)
        # if (mpi_rank == 0 ):
        #     print(res)

        if (res[0] == "init"):

            (ker_lcm, X, Y, maxtries,jitter) = res[1]
            mb = min(mb, max(1,min(X.shape[0]//nprow, X.shape[0]//npcol)))   # YL: mb <=32 doesn't seem reasonable, comment this line out ?
            # # print('mb',mb,'nprow',nprow,'npcol',npcol)
            cliblcm.initialize.restype = POINTER(fun_jac_struct)
            z = cliblcm.initialize (\
                    c_int(ker_lcm.input_dim - 1),\
                    c_int(ker_lcm.num_outputs),\
                    c_int(ker_lcm.Q),\
                    c_int(X.shape[0]),\
                    X.ctypes.data_as(POINTER(c_double)),\
                    Y.ctypes.data_as(POINTER(c_double)),\
                    c_int(mb),\
                    c_int(maxtries),\
                    c_double(jitter),\
                    c_int(nprow),\
                    c_int(npcol),\
                    c_mpi_comm_t.from_address(mpi4py.MPI._addressof(mpi4py.MPI.COMM_WORLD)))

        elif (res[0] == "fun_jac"):
            x2 = res[1]
            gradients = np.zeros(len(ker_lcm.theta) + len(ker_lcm.var) + len(ker_lcm.kappa) + len(ker_lcm.sigma) + len(ker_lcm.WS))
            cliblcm.fun_jac.restype = c_double

            # res = mpi_comm.bcast(None, root=mpi4py.MPI.ROOT)
            # print('check',res)

            neg_log_marginal_likelihood = cliblcm.fun_jac ( x2.ctypes.data_as(POINTER(c_double)), z, gradients.ctypes.data_as(POINTER(c_double)) )
            if (mpi_rank == 0):
                mpi_comm.send((neg_log_marginal_likelihood, gradients), dest=0)

        elif (res[0] == "end"):

            cond = False
            cliblcm.finalize(z)
            mpi_comm.Disconnect()




from setuptools import setup
import os
import io

NAME = 'GPTune'
DESCRIPTION = ''

here = os.path.abspath(os.path.dirname(__file__))

# Load the package's __version__.py module as a dictionary.
about = dict()
with open(os.path.join(here, NAME, '__version__.py')) as f:
    exec(f.read(), about)

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

REQUIRED=[]

EXTRAS = {
    'all': [
        'Sphinx>=1.8.2',
        'sphinx_rtd_theme',
        'numpy',
        'joblib',
        'scikit-learn <= 0.22',
        'scipy',
        'pyaml',
        'matplotlib',
        'GPy',
        'openturns',
        'lhsmdu',
        'ipyparallel',
        'opentuner',
        'hpbandster',
        'pygmo',
        'filelock', 
        'requests',           
    ],
}

setup(
  name=NAME,
  version=about['__version__'],
  license='BSD',
  author="Yang Liu",
  author_email='liuyangzhuan@lbl.gov',
  description=DESCRIPTION,
  packages=['GPTune'],
  url='https://github.com/gptune/GPTune',
  install_requires=REQUIRED,
  extras_require=EXTRAS,
)